<?php

/**
 * Menu callback, redirect project committers.
 */
function drupalorg_project_committers_redirect(stdClass $node) {
  drupal_goto(versioncontrol_project_repository_load($node->nid)->getUrlHandler()->getRepositoryViewUrl() . '/-/graphs/HEAD');
}

/**
 * Menu callback, redirect project commits.
 */
function drupalorg_project_commits_redirect(stdClass $node, $feed = FALSE) {
  $url = versioncontrol_project_repository_load($node->nid)->getUrlHandler()->getRepositoryViewUrl();
  if ($feed) {
    $url .= '.atom';
  }
  else {
    $url .= '/-/commits';
  }
  drupal_goto($url);
}

/**
 * Menu callback, redirect user commits.
 */
function drupalorg_user_commits_redirect(stdClass $account, $feed = FALSE) {
  if ($feed) {
    drupal_goto(variable_get('versioncontrol_gitlab_url') . '/' . $account->git_username . '.atom');
  }
  else {
    drupal_goto(variable_get('versioncontrol_gitlab_url') . '/users/' . $account->git_username . '/activity');
  }
}

/**
 * Redirect an individual commit page.
 *
 * @param VersioncontrolRepository $repository
 * @param string $revision
 */
function drupalorg_commit_redirect(VersioncontrolRepository $repository, $revision) {
  drupal_goto($repository->getUrlHandler()->getCommitViewUrl($revision));
}

/**
 * Menu callback, redirect to the appropriate place to report a security issue
 * for a project.
 */
function drupalorg_project_security_redirect(stdClass $node) {
  $wrapper = entity_metadata_wrapper('node', $node);
  $project = $wrapper->field_project_machine_name->value();

  if ($wrapper->field_security_advisory_coverage->value() === 'covered' && drupalorg_project_has_full_release($node)) {
    drupal_goto('https://security.drupal.org/node/add/project-issue/' . $project);
  }
  else {
    drupal_set_message(t('This project is not covered by the Drupal Security Team’s advisory policy. Security issues do not need to be privately reported for the %title project.', ['%title' => $node->title]));
    drupal_goto('node/add/project-issue/' . $project, ['query' => ['tags' => 'security', 'priorities' => '400', 'categories' => 'bug']]);
  }
}

/**
 * Menu callback, project browsing pages.
 */
function drupalorg_project_browse_page($project_type) {
  // Default Solr params.
  $params = [
    'q' => isset($_GET['text']) && drupal_validate_utf8($_GET['text']) ? $_GET['text'] : '',
    'qf' => 'content',
    'fl' => [
      'id',
      'entity_type',
      'entity_id',
      'label',
      'content',
      'is_comment_count',
      'bundle',
      'ds_created',
      'ds_changed',
      'url',
      'is_uid',
      'ss_name',
      'iss_project_release_usage',
      'ds_project_latest_release',
      'hash',
    ],
    'rows' => variable_get('apachesolr_rows', 10),
    'facet' => 'true',
    'facet.mincount' => 1,
    'facet.sort' => 'true',
    'facet.field' => [
      'bundle',
    ],
    'facet.limit' => 200,
  ];
  if (isset($_GET['page']) && ctype_digit($_GET['page'])) {
    $params['start'] = $_GET['page'] * $params['rows'];
  }

  // This is the object that does the communication with the solr server.
  $env_id = apachesolr_default_environment();
  $solr = apachesolr_get_solr($env_id);
  if (isset($_GET['solrsort'])) {
    $solrsort = $_GET['solrsort'];
  }
  else {
    $solrsort = 'score desc';
  }
  $query = new SolrBaseQuery($env_id, $solr, $params, $solrsort, $_GET['q']);
  if (is_null($query)) {
    throw new Exception(t('Could not construct a Solr query.'));
  }

  // We add addFilter() parameters here to include all the constant filters for
  // the query -- project nodes of the given top-level type that have releases
  // (if project_release is enabled). We use addFilter() rather than
  // $params['fq'] so that our filters are correctly passed to anything that
  // uses our cached query.
  $query->addFilter('bundle', $project_type->type);

  // Apply sort value.
  list($sort_field, $sort_direction) = explode(' ', $solrsort);
  $query->setSolrsort($sort_field, $sort_direction);

  // Allow modules to alter the query prior to statically caching it.
  // This can e.g. be used to add available sorts.
  foreach (module_implements('apachesolr_query_prepare') as $module) {
    $function_name = $module . '_apachesolr_query_prepare';
    $function_name($query);
  }
  apachesolr_search_add_boost_params($query);

  // Cache the built query. Since all the built queries go through
  // this process, all the hook_invocations will happen later.
  apachesolr_current_query($env_id, $query);
  try {
    $response = $query->search($params['q']);

    // The response is cached so that it is accessible to the blocks and
    // anything else that needs it beyond the initial search.
    apachesolr_static_response_cache($query->getSearcher(), $response);
    apachesolr_has_searched($env_id, TRUE);

    $nids = [];
    foreach (apachesolr_search_process_response($response, $query) as $result) {
      $nids[] = $result['node']->entity_id;
    }
    $results = [];
    foreach (node_load_multiple($nids) as $node) {
      $body_no_tags = !empty($node->body[LANGUAGE_NONE]) ? strip_tags($node->body[LANGUAGE_NONE][0]['value']) : '';
      $body_no_breaks = trim(preg_replace('/[\t\n\r\s]+/', ' ', $body_no_tags));
      $body_truncated = truncate_utf8($body_no_breaks, 200, TRUE, TRUE, 1);
      $body = check_plain($body_truncated);
      $terms = _drupalorg_get_node_categories_as_links($node);
      if (!empty($terms)) {
        $body .= '<br><span class="terms description">' . t('Categories: ') . implode(', ', $terms) . '</span>';
      }
      $results[] = [
        'title' => [
          '#markup' => '<h2>' . _drupalorg_get_repository_logo($node) . l($node->title, 'node/' . $node->nid) . '</h2>',
        ],
        'body' => [
          '#markup' => '<p>' . $body . '</p>',
        ],
      ];
    }
  }
  catch (\Exception $e) {
    drupal_set_message(t('Error in searching: @message', [
      '@message' => $e->getMessage(),
    ]), 'error');
  }

  $form_state = [
    'method' => 'get',
    'always_process' => TRUE,
    'no_redirect' => TRUE,
    'build_info' => ['args' => [$project_type]],
  ];
  $form = drupal_build_form('drupalorg_browse_projects_form', $form_state);
  unset($form['#build_id'], $form['form_build_id'], $form['form_id']);

  return [
    'drupalorg_browse_projects_form' => $form,
    'description' => ['#markup' => '<p>' . filter_xss($project_type->description) . '</p>'],
    'results' => $results,
    'pager' => ['#markup' => theme('pager')],
  ];
}

/**
 * Extract categories from a node and return them as links.
 */
function _drupalorg_get_node_categories_as_links($node) {
  $links = [];
  $categories = [];

  if (!empty($node->taxonomy_vocabulary_3)) {
    foreach ($node->taxonomy_vocabulary_3[LANGUAGE_NONE] as $key => $value) {
      $categories[] = $value['tid'];
    }
  }

  if (!empty($categories)) {
    foreach (taxonomy_term_load_multiple($categories) as $term) {
      $query = $_REQUEST;
      $query['f'][3] = 'im_vid_3:' . $term->tid;
      $links[] = l($term->name, menu_get_item()['path'], ['query' => $query]);
    }
  }

  return $links;
}

/**
 * Page callback, shortcut to get to issue pages.
 */
function drupalorg_issue_redirect_page($nid) {
  drupal_goto('node/'. $nid);
}

/**
 * Page callback, redirect to newsletter signup.
 */
function drupalorg_subscribe_redirect_page() {
  if (user_is_logged_in()) {
    drupal_goto('user/' . $GLOBALS['user']->uid . '/edit', ['fragment' => 'edit-field-subscribe-to']);
  }
  else {
    drupal_goto('association/newsletter-subscriptions');
  }
}

/**
 * Form callback, redirect to Slack invite URL.
 */
function drupalorg_join_slack_form() {
  return [
    'agreement' => [
      '#type' => 'checkbox',
      '#title' => t('As with all Drupal community spaces, by participating in Drupal Slack you agree to adhere to the <a href="/dcoc">Drupal Code of Conduct</a>'),
      '#required' => TRUE,
    ],
    'actions' => [
      '#type' => 'actions',
      'join' => [
        '#type' => 'submit',
        '#value' => t('Join'),
      ],
    ],
  ];
}

/**
 * Form submit callback for drupalorg_join_slack_form().
 */
function drupalorg_join_slack_form_submit($form, &$form_state) {
  $form_state['redirect'] = variable_get('drupalorg_slack_invite');
}

/**
 * Page callback, redirect to a sponsor.
 */
function drupalorg_sponsor_redirect_page($sponsor) {
  $country = drupalorg_crosssite_get_and_vary_header('GeoIP-country');
  switch ($sponsor) {
    case '1and1-shared':
      switch ($country) {
        case 'DE':
          drupal_goto('https://www.ionos.de/hosting/drupal-hosting?ac=OM.PU.PUo63K403739T7073a');
          return;

        case 'AT':
          drupal_goto('https://www.ionos.at/hosting/drupal-hosting?ac=OM.AA.AAo63K403756T7073a');
          return;

        case 'GB':
          drupal_goto('https://www.ionos.co.uk/hosting/drupal-hosting?ac=OM.UK.UKo63K403741T7073a');
          return;

        case 'FR':
          drupal_goto('https://www.ionos.fr/hebergement/hebergement-web?ac=OM.FR.FRo63K403742T7073a');
          return;

        case 'ES':
          drupal_goto('https://www.ionos.es/alojamiento/alojamiento-web?ac=OM.WE.WEo63K403743T7073a');
          return;

        case 'CA':
          drupal_goto('https://www.ionos.ca/hosting/drupal-hosting?ac=OM.CA.CAo63K403744T7073a');
          return;

        case 'MX':
          drupal_goto('https://www.ionos.mx/alojamiento/web-hosting?ac=OM.MB.MBo63K403745T7073a');
          return;

        case 'IT':
          drupal_goto('https://www.ionos.it/hosting/web-hosting?ac=OM.IA.IAo63K403746T7073a');
          return;

        default:
          drupal_goto('https://www.ionos.com/hosting/drupal-hosting?ac=OM.US.USo63K403740T7073a');
          return;
      }

    case '1and1-cloud':
      switch ($country) {
        case 'DE':
          drupal_goto('https://hosting.1und1.de/cloud-app-center/drupal-download?ac=OM.PU.PUo63K403739T7073a');
          return;

        case 'AT':
          drupal_goto('https://www.1und1.at/cloud-app-center/drupal-download?ac=OM.AA.AAo63K403756T7073a');
          return;

        case 'GB':
          drupal_goto('https://www.1and1.co.uk/cloud-app-centre/drupal-download?ac=OM.UK.UKo63K403741T7073a');
          return;

        case 'FR':
          drupal_goto('https://www.1and1.fr/cloud-app-center/drupal-download?ac=OM.FR.FRo63K403742T7073a');
          return;

        case 'ES':
          drupal_goto('https://www.1and1.es/cloud-app-center/drupal-descarga?ac=OM.WE.WEo63K403743T7073a');
          return;

        case 'CA':
          drupal_goto('https://www.1and1.ca/cloud-app-center/drupal-download?ac=OM.CA.CAo63K403744T7073a');
          return;

        case 'MX':
          drupal_goto('https://www.1and1.mx/cloud-app-center/drupal-descarga?ac=OM.MB.MBo63K403745T7073a');
          return;

        case 'IT':
          drupal_goto('https://www.1and1.it/cloud-app-center/drupal-download?ac=OM.IA.IAo63K403746T7073a');
          return;

        default:
          drupal_goto('https://www.1and1.com/cloud-app-center/drupal-download?ac=OM.US.USo63K403740T7073a');
          return;
      }

    case '1and1-vps':
      switch ($country) {
        case 'DE':
          drupal_goto('https://hosting.1und1.de/vserver?ac=OM.PU.PUo63K403739T7073a');
          return;

        case 'AT':
          drupal_goto('https://www.1und1.at/vserver?ac=OM.AA.AAo63K403756T7073a');
          return;

        case 'GB':
          drupal_goto('https://www.1and1.co.uk/virtual-server?ac=OM.UK.UKo63K403741T7073a');
          return;

        case 'FR':
          drupal_goto('https://www.1and1.fr/serveurs-virtuels?ac=OM.FR.FRo63K403742T7073a');
          return;

        case 'ES':
          drupal_goto('https://www.1and1.es/servidores-virtuales?ac=OM.WE.WEo63K403743T7073a');
          return;

        case 'CA':
          drupal_goto('https://www.1and1.ca/vps-hosting?ac=OM.CA.CAo63K403744T7073a');
          return;

        case 'MX':
          drupal_goto('https://www.1and1.mx/servidores-virtuales?ac=OM.MB.MBo63K403745T7073a');
          return;

        case 'IT':
          drupal_goto('https://www.1and1.it/server-virtuali?ac=OM.IA.IAo63K403746T7073a');
          return;

        default:
          drupal_goto('https://www.1and1.com/vps-hosting?ac=OM.US.USo63K403740T7073a');
          return;
      }

    case '1and1-dedicated':
      switch ($country) {
        case 'DE':
          drupal_goto('https://hosting.1und1.de/dedicated-server-tarife?ac=OM.PU.PUo63K403739T7073a');
          return;

        case 'AT':
          drupal_goto('https://www.1und1.at/dedicated-server-tarife?ac=OM.AA.AAo63K403756T7073a');
          return;

        case 'GB':
          drupal_goto('https://www.1and1.co.uk/server-dedicated-tariff?ac=OM.UK.UKo63K403741T7073a');
          return;

        case 'FR':
          drupal_goto('https://www.1and1.fr/server-dedicated-tariff?ac=OM.FR.FRo63K403742T7073a');
          return;

        case 'ES':
          drupal_goto('https://www.1and1.es/server-dedicated-tariff?ac=OM.WE.WEo63K403743T7073');
          return;

        case 'CA':
          drupal_goto('https://www.1and1.ca/server-dedicated-tariff?ac=OM.CA.CAo63K403744T7073a');
          return;

        case 'MX':
          drupal_goto('https://www.1and1.mx/server-dedicated-tariff?ac=OM.MB.MBo63K403745T7073a');
          return;

        case 'IT':
          drupal_goto('https://www.1and1.it/server-dedicated-tariff?ac=OM.IA.IAo63K403746T7073a');
          return;

        default:
          drupal_goto('https://www.1and1.com/server-dedicated-tariff?ac=OM.US.USo63K403740T7073a');
          return;
      }

    case '1and1-try-drupal':
      switch ($country) {
        case 'DE':
          drupal_goto('https://www.ionos.de/hosting/drupal-cloud-hosting?ac=OM.PU.PUo63K404685T7073a');
          return;

        case 'AT':
          drupal_goto('https://www.ionos.at/hosting/drupal-cloud-hosting?ac=OM.AA.AAo63K404687T7073a');
          return;

        case 'GB':
          drupal_goto('https://www.ionos.co.uk/hosting/drupal-cloud-hosting?ac=OM.UK.UKo63K404688T7073a');
          return;

        case 'FR':
          drupal_goto('https://www.ionos.fr/hebergement/drupal-cloud-hosting?ac=OM.FR.FRo63K404689T7073a');
          return;

        case 'ES':
          drupal_goto('https://www.ionos.es/alojamiento/drupal-cloud-hosting?ac=OM.WE.WEo63K404690T7073a');
          return;

        case 'CA':
          drupal_goto('https://www.ionos.ca/hosting/drupal-cloud-hosting?ac=OM.CA.CAo63K404691T7073a');
          return;

        case 'MX':
          drupal_goto('https://www.ionos.mx/alojamiento/drupal-cloud-hosting?ac=OM.MB.MBo63K404692T7073a');
          return;

        case 'IT':
          drupal_goto('https://www.ionos.mx/alojamiento/drupal-cloud-hosting?ac=OM.IA.IAo63K404693T7073a');
          return;

        default:
          drupal_goto('https://www.ionos.com/hosting/drupal-cloud-hosting?ac=OM.US.USo63K404686T7073a');
          return;
      }

    case 'debug':
      return t('Country: %country', ['%country' => $country]);
  }
}

/**
 * Page callback, current user becomes maintainer of a guide.
 */
function drupalorg_maintain_guide($node) {
  // Make current user the author, OG auotmatically adds maintainership.
  $node->uid = $GLOBALS['user']->uid;
  node_save($node);

  // Remove System Message placeholder maintainer.
  og_get_membership('node', $node->nid, 'user', 180064)->delete();

  drupal_set_message(t('You now maintain this guide! <a href="!url">Learn about maintaining documentation</a>', ['!url' => url('drupalorg/docs/content/documentation', ['fragment' => 'maintain'])]));

  drupal_goto('node/' . $node->nid);
}

/**
 * Redirect to a user’s edit page.
 */
function drupalorg_user_edit_redirect($fragment = NULL) {
  $options = [];
  if (!is_null($fragment)) {
    $options['fragment'] = $fragment;
  }
  drupal_goto('user/' . $GLOBALS['user']->uid . '/edit', $options);
}

/**
 * Redirect to a user’s membership page.
 */
function drupalorg_my_membership_redirect() {
  drupal_goto('user/' . $GLOBALS['user']->uid . '/membership');
}

/**
 * Membership summary & update form.
 */
function drupalorg_membership_page(stdClass $account) {
  if (!empty($account->field_da_ind_membership[LANGUAGE_NONE][0]['value'])) {
    return '<p>' . l(t('Manage your membership'), 'https://www.classy.org/profile', [
      'attributes' => ['class' => 'action-button'],
      'external' => TRUE,
    ]) . '</p><p>' . t('<em><strong>Current or expired members:</strong>
To upgrade or renew your membership, visit <a href="https://www.classy.org/sso">Classy.org</a> and sign in to your account to make any changes. If you have not already set up an account with Classy, please visit <a href="https://www.classy.org/sso">Classy.org</a> and select “Create an account” using the same email address you gave when you signed up for your membership.</em></p>');
  }
  else {
    return '<p>' . l(t('Join today'), 'https://www.classy.org/give/386338', [
      'attributes' => ['class' => 'action-button'],
      'external' => TRUE,
    ]) . '</p>';
  }
}

/**
 * Menu callback to redirect legacy Drupal service provider listings.
 */
function drupalorg_redirect_legacy_services($service = NULL, $sector = NULL, $location = NULL) {
  $query = [];

  if ($service !== 'all') {
    $terms = taxonomy_get_term_by_name($service, 'vocabulary_48');
    $service_tid = key($terms);
    if (!empty($service_tid)) {
      $query['services'] = [$service_tid];
    }
  }

  if ($sector !== 'all') {
    $terms = taxonomy_get_term_by_name($sector, 'vocabulary_50');
    $sector_tid = key($terms);
    if (!empty($sector_tid)) {
      $query['sectors'] = [$sector_tid];
    }
  }

  if ($location !== 'all') {
    $terms = taxonomy_get_term_by_name($location, 'vocabulary_52');
    $location_tid = key($terms);
    if (!empty($location_tid)) {
      $query['offices'] = [$location_tid];
    }
  }

  drupal_goto('drupal-services', ['query' => $query]);
}

/**
 * Menu callback, go to the current D8 release page.
 */
function drupalorg_download_redirect() {
  drupal_goto('node/' . drupalorg_get_latest_supported_stable_release(DRUPALORG_CORE_NID));
}

/**
 * Menu callback, redirect to most recent Drupal tarball or zip file.
 */
function drupalorg_latest_download($extension) {
  $project_nid = DRUPALORG_CORE_NID;

  if ($extension === 'cms') {
    $extension = 'zip';
    $project_nid = project_get_nid_from_machinename('cms');
  }

  if (!in_array($extension, ['tar.gz', 'zip'])) {
    return MENU_NOT_FOUND;
  }

  $release_node = node_load(drupalorg_get_latest_supported_stable_release($project_nid));
  foreach ($release_node->field_release_files[LANGUAGE_NONE] as $delta => $row) {
    $field_collection_item_id = $release_node->field_release_files[LANGUAGE_NONE][$delta]['value'];
    $field_collection_item = field_collection_item_load($field_collection_item_id);
    $uri = $field_collection_item->field_release_file[LANGUAGE_NONE][0]['uri'];
    if (preg_match('/\.' . preg_quote($extension, '/') . '$/', $uri)) {
      drupal_goto(project_release_add_download_base($uri, variable_get('project_release_download_base_' . $release_node->type, '')));
    }
  }

  return MENU_NOT_FOUND;
}

/**
 * Get the latest supported stable release nid for a project.
 */
function drupalorg_get_latest_supported_stable_release($project_nid) {
  // Get “recommended” releases for supported branches. Recommended is a stable
  // version, if available.
  $recommended_releases = db_query('SELECT branch, recommended_release FROM {project_release_supported_versions} WHERE nid = :nid AND supported = 1', [':nid' => $project_nid])->fetchAllKeyed();
  if (count($recommended_releases) === 0) {
    return FALSE;
  }
  // Check if those releases have extra versions, like alpha42 or dev.
  $query = new EntityFieldQuery();
  $with_extra_release = $query->entityCondition('entity_type', 'node')
    ->entityCondition('bundle', project_release_release_node_types())
    ->propertyCondition('status', NODE_PUBLISHED)
    ->propertyCondition('nid', $recommended_releases)
    ->fieldCondition('field_release_version_extra', 'value', '', '!=')
    ->execute();
  if (isset($with_extra_release['node'])) {
    $recommended_releases = array_diff($recommended_releases, array_keys($with_extra_release['node']));
  }
  uksort($recommended_releases, 'project_release_version_compare');

  return end($recommended_releases);
}

/**
 * Menu callback, 403 for unsupported GitLab operations.
 */
function drupalorg_git_error() {
  drupal_add_http_header('Status', '403 Access Denied');
  return t('<p>You have attempted an action in GitLab that Drupal.org does not currently support.</p>'
    . '<p>Please see our <a href="/docs/develop/git/using-git-to-contribute-to-drupal/gitlab-frequently-asked-questions#s-im-receiving-an-error-message-or-an-unsupported-operation-warning">GitLab documentation</a> for more information.</p>');
}

/**
 * Menu callback for GitLab’s /etc/gitlab-shell/update.d/update.
 *
 * @param int $gitlab_project_id
 *   A GitLab project ID.
 * @param string $ref
 *   'refs/heads/branch-name' or 'refs/tags/tag-name'
 * @param string $old_commit
 *   The commit hash this label is being moved from.
 * @param string $new_commit
 *   The commit hash this label will point at, if allowed.
 */
function drupalorg_gitlab_validate($gitlab_project_id, $ref, $old_commit, $new_commit) {
  $msg = 'OK';

  try {
    $repository = versioncontrol_repository_load(versioncontrol_gitlab_get_repo_id($gitlab_project_id));
    $project = node_load($repository->project_nid);

    if ($project->status == NODE_NOT_PUBLISHED) {
      drupal_json_output(t('This project is not published.'));
      return;
    }

    $project_wrapper = entity_metadata_wrapper('node', $project);
    $auth_data = versioncontrol_project_get_auth_data($project_wrapper->field_project_machine_name->value());

    list(, $type, $name) = explode('/', urldecode($ref), 3);
    $is_branch = $type === 'heads';
    if (array_search($name, $auth_data['protected_labels'][$is_branch ? 'branches' : 'tags'], TRUE) !== FALSE) {
      if ($is_branch) {
        // Disallow the deletion of branches or tags which already have associated releases.
        if ($new_commit == GIT_NULL_REV) {
          $msg = t('The branch !name is tied to a development release on Drupal.org, and cannot be deleted.', ['!name' => $name]);
        }
        elseif (!$repository->isFastForward($old_commit, $new_commit)) {
          $msg = t('The branch !name is tied to a development release on Drupal.org, and force pushes are not allowed.', ['!name' => $name]);
        }
      }
      else {
        $msg = t('The tag !name is tied to a release on Drupal.org, and cannot be relocated or deleted.', ['!name' => $name]);
      }
    }
  }
  catch (Exception $e) {
    watchdog('drupalorg_gitlab_validate', 'Exception in push validation, GitLab project @gitlab_project_id, ref @ref, @old_commit → @new_commit: @message', [
      '@gitlab_project_id' => $gitlab_project_id,
      '@ref' => $ref,
      '@old_commit' => $old_commit,
      '@new_commit' => $new_commit,
      '@message' => $e->getMessage(),
    ], WATCHDOG_ERROR);
    $msg = t('Drupal.org pre-commit check failed. Please try again in a moment. (If this issue persists, please report to https://www.drupal.org/project/issues/infrastructure)');
  }

  drupal_json_output($msg);
}

/**
 * Menu callback, post-receive hook for issue fork repositories.
 */
function drupalorg_issue_fork_post_receive() {
  if (!isset($_POST['payload']) || !isset($_POST['hash'])) {
    return MENU_ACCESS_DENIED;
  }
  if ($_POST['hash'] != sha1(variable_get('versioncontrol_gitlab_push_key') . $_POST['payload'])) {
    return MENU_ACCESS_DENIED;
  }
  $payload = json_decode($_POST['payload']);
  foreach (explode("\n", $payload->data, -1) as $line) {
    list($from_id, $to_id, $ref) = explode(' ', $line);

    // Mark branch as active.
    if (preg_match('#^refs/heads/(?<branch>.*)$#', $ref, $match)) {
      db_merge('drupalorg_issue_fork_branches')
        ->key([
          'gitlab_project_id' => $payload->repo_id,
          'branch' => $match['branch'],
        ])
        ->fields([
          'last_activity' => REQUEST_TIME,
        ])
        ->execute();
    }

    // For issue credit, record how many times someone has pushed to an issue
    // fork. If the committer or author has not posted a comment to the issue
    // yet, post one on behalf of them.
    if (!preg_match('#^refs/tags/#', $ref)) {
      $result = (new EntityFieldQuery())->entityCondition('entity_type', 'user')
        ->propertyCondition('git_username', $payload->git_username)
        ->range(0, 1)->execute();
      if (isset($result['user'])) {
        $uid = array_keys($result['user'])[0];
        $issue_fork = entity_load_single('drupalorg_issue_fork', $payload->repo_id);
        db_merge('drupalorg_issue_fork_maintainers')
          ->insertFields(['pushes' => 1])
          ->key([
            'gitlab_project_id' => $issue_fork->gitlab_project_id,
            'uid' => $uid,
          ])
          ->expression('pushes', 'pushes + 1')
          ->fields(['last_push' => REQUEST_TIME])
          ->execute();
        $result = (new EntityFieldQuery())->entityCondition('entity_type', 'comment')
          ->propertyCondition('uid', $uid)
          ->propertyCondition('nid', $issue_fork->nid)
          ->range(0, 1)->execute();
        if (!isset($result['comment'])) {
          drupalorg_post_issue_comment_for_credit($issue_fork->nid, user_load($uid), '!name made their first commit to this issue’s fork.');
        }
      }
      else {
        watchdog('drupalorg_issue_fork', 'Push not matched to a Drupal.org user. <code>@payload</code>', [
          '@payload' => print_r($payload, TRUE),
        ], WATCHDOG_ERROR);
      }
    }
  }
}

/**
 * Menu callback, handle GitLab merge request event webhook.
 */
function drupalorg_issue_fork_merge_request_event() {
  $body = json_decode(file_get_contents('php://input'));

  if (isset($body->object_kind) && $body->object_kind === 'merge_request' && !in_array($body->object_attributes->target->namespace, ['issue', 'Issue forks']) && ($issue_fork = entity_load_single('drupalorg_issue_fork', $body->object_attributes->source_project_id))) {
    $fields = [
      'last_activity' => REQUEST_TIME,
      'merge_request_iid' => $body->object_attributes->iid,
      'merge_request_target_branch' => $body->object_attributes->target_branch,
    ];
    switch ($body->object_attributes->action) {
      case 'close':
        $fields['status'] = 'hidden';
        break;

      case 'open':
      case 'reopen':
        $fields['status'] = 'active';
        break;
    }
    db_merge('drupalorg_issue_fork_branches')
      ->key([
        'gitlab_project_id' => $issue_fork->gitlab_project_id,
        'branch' => $body->object_attributes->source_branch,
        'merge_request_iid' => [0, $body->object_attributes->iid],
      ])
      ->fields($fields)
      ->expression('merge_request_initial_commit', 'coalesce(merge_request_initial_commit, :commit)', [
        ':commit' => $body->object_attributes->last_commit->id,
      ])
      ->execute();

    DrupalQueue::get('drupalorg_issue_forks')->createItem([
      'fork' => $issue_fork,
      'task' => 'receiveMergeRequestEvent',
      'arguments' => [$body],
    ]);
  }

  drupal_add_http_header('Content-Type', 'text/plain');
  echo '👋';
}

/**
 * Menu callback, handle GitLab repository update webhook.
 */
function drupalorg_repository_update_event() {
  $body = json_decode(file_get_contents('php://input'));

  if (in_array($body->event_name, ['user_add_to_team', 'user_remove_from_team'])) {
    // Find user account.
    $uid = db_query('SELECT uid FROM {versioncontrol_gitlab_users} WHERE gitlab_user_id = :gitlab_user_id', [
      ':gitlab_user_id' => $body->user_id,
    ])->fetchField();
    if (empty($uid)) {
      watchdog('drupalorg_gitlab', 'Repository update event not matched to a Drupal.org user. <code>@payload</code>', [
        '@payload' => print_r($body, TRUE),
      ], WATCHDOG_ERROR);
      return;
    }

    if ($repository_id = versioncontrol_gitlab_get_repo_id($body->project_id)) {
      // A project’s repository is being updated.
      $project_node = versioncontrol_project_project_load($repository_id);
      switch ($body->event_name) {
        case 'user_add_to_team':
          drupalorg_project_maintainer_save($project_node->nid, $uid, ['write to vcs' => TRUE]);
          break;

        case 'user_remove_from_team':
          if ($uid == $project_node->uid) {
            // Default to Unsupported Projects user.
            $project_node->uid = 291168;
            // Look for another person with “administer maintainers” access.
            foreach ($project_node->project['maintainers'] as $maintainer_uid => $info) {
              if ($maintainer_uid != $uid && !empty($info['permissions']['administer maintainers'])) {
                $project_node->uid = $maintainer_uid;
                break;
              }
            }
            $project_node->revision = 1;
            $project_node->log = t('@name left the project’s repository.', [
              '@name' => format_username(user_load($uid)),
            ]);
            node_save($project_node);
          }
          drupalorg_project_maintainer_remove($project_node->nid, $uid);
          break;
      }
    }
    elseif ($issue_fork = entity_load_single('drupalorg_issue_fork', $body->project_id)) {
      // An issue fork repository is being updated.
      db_merge('drupalorg_issue_fork_maintainers')
        ->key([
          'gitlab_project_id' => $issue_fork->gitlab_project_id,
          'uid' => $uid,
        ])
        ->fields([
          'current_member' => (int) ($body->event_name === 'user_add_to_team'),
        ])
        ->execute();
    }
  }
  elseif ($body->event_name === 'repository_update') {
    // There was a push to the project’s repository.
    watchdog('drupalorg_gitlab', 'Purging <code>@key</code>', [
      '@key' => $body->project->name,
    ], WATCHDOG_DEBUG);
    drupalorg_fastly_purge_key($body->project->name);

    // Project logos are cached permanently, so clear it in case there is a new one.
    cache_clear_all('drupalorg_repository_logo_' . $body->project_id, 'cache');

    if ($body->project->namespace === 'project' && ($repository_id = versioncontrol_gitlab_get_repo_id($body->project_id))) {
      $project_node = versioncontrol_project_project_load($repository_id);

      foreach ($body->refs as $ref) {
        if (preg_match('#^refs/heads/(?<branch_name>.*)$#', $ref, $match)) {
          // Package dev releases.
          if (($release = project_release_get_release_for_vcs_label($project_node->nid, $match['branch_name']))) {
            drupalorg_create_packaging_job($release);
          }

          // Queue to see if we need to add comments on Drupal.org issues.
          try {
            DrupalQueue::get('drupalorg_commits_comments')->createItem([
              'event_details' => $body,
              'project_nid' => $project_node->nid,
              'ref' => $ref,
            ]);
          }
          catch (Exception $e) {
            watchdog('drupalorg_gitlab', 'Failed to enqueue commit comment job. Error was !error', [
              '!error' => $e->getMessage(),
            ], WATCHDOG_ERROR);
          }
        }
      }
    }
  }

  drupal_add_http_header('Content-Type', 'text/plain');
  echo '👋';
}

/**
 * Menu callback, queues in place update “quasi diff” generation.
 */
function drupalorg_queue_in_place_update($project, $filename) {
  // Make sure project in filename matches project path argument.
  $project_wrapper = entity_metadata_wrapper('node', $project);
  if ($project_wrapper->field_project_machine_name->value() !== $filename['project']) {
    return MENU_NOT_FOUND;
  }

  // Make sure releases exist.
  $tagged_releases_only = [[
    'field' => 'field_release_build_type',
    'column' => 'value',
    'value' => 'static',
  ]];
  if (!($from_release_nid = project_release_exists($project->nid, $filename['from'], $tagged_releases_only))) {
    return MENU_NOT_FOUND;
  }
  if (!($to_release_nid = project_release_exists($project->nid, $filename['to'], $tagged_releases_only))) {
    return MENU_NOT_FOUND;
  }

  $key = [
    'from_release_nid' => $from_release_nid,
    'to_release_nid' => $to_release_nid,
  ];

  // Record the request.
  $status = db_merge('drupalorg_in_place_updates')
    ->insertFields([
      'requested' => REQUEST_TIME,
    ])
    ->key($key)
    ->execute();

  // Queue if this is the first request.
  if ($status === MergeQuery::STATUS_INSERT) {
    DrupalQueue::get('drupalorg_in_place_updates')->createItem($key);
  }

  // Return message while queued.
  header('Retry-After: 10', TRUE, 429);
  drupal_set_message(t('This update is being generated and is not available yet.'), 'warning');
  return '';
}

/**
 * Menu callback, redirect old case study view links.
 */
function drupalorg_legacy_case_study_redirect($case_study_type, $term_name) {
  if ($case_study_type === 'community') {
    if (1 === preg_match('/drupal-[0-9]+\.x/', $term_name)) {
      $term_id = key(taxonomy_get_term_by_name(str_replace('-', ' ', $term_name), 'vocabulary_5'));
      drupal_goto('case-studies/community', ['query' => ['version' => [$term_id]]]);
    }
    else {
      $query = db_select('taxonomy_term_data', 'td');
      $query->addTag('taxonomy_term_access');
      $query->leftJoin('taxonomy_vocabulary', 'tv', 'td.vid = tv.vid');
      $query->fields('td');
      $query->fields('tv', array('machine_name'));
      $query->where("replace(td.name, ' ', '-') = :name", array(':name' => $term_name));
      $vocabularies = array('vocabulary_50', 'vocabulary_5');
      if (!empty($vocabularies)) {
        $query->condition('tv.machine_name', $vocabularies);
      }
      if ($term = $query->execute()->fetchObject()) {
        drupal_goto('/case-studies/community', ['query' => ['sectors' => [$term->tid]]]);
      }
      else {
        drupal_goto('/case-studies/community');
      }
    }
  }
  elseif ($case_study_type === 'featured') {
    if (1 === preg_match('/drupal-[0-9]+\.x/', $term_name)) {
      $term_id = key(taxonomy_get_term_by_name(str_replace('-', ' ', $term_name), 'vocabulary_5'));
      drupal_goto('case-studies', ['query' => ['version' => [$term_id]]]);
    }
    else {
      $query = db_select('taxonomy_term_data', 'td');
      $query->addTag('taxonomy_term_access');
      $query->leftJoin('taxonomy_vocabulary', 'tv', 'td.vid = tv.vid');
      $query->fields('td');
      $query->fields('tv', array('machine_name'));
      $query->where("replace(td.name, ' ', '-') = :name", array(':name' => $term_name));
      $vocabularies = array('vocabulary_50', 'vocabulary_5');
      if (!empty($vocabularies)) {
        $query->condition('tv.machine_name', $vocabularies);
      }
      if ($term = $query->execute()->fetchObject()) {
        drupal_goto('/case-studies', ['query' => ['sectors' => [$term->tid]]]);
      }
      else {
        drupal_goto('/case-studies');
      }
    }
  }
}

/**
 * Fetches Tugboat previews for an issue fork.
 *
 * @param string $issue_fork
 *   The issue fork name.
 */
function drupalorg_issue_fork_tugboat_previews($issue_fork) {
  drupal_json_output($issue_fork->getTugboatPreviews());
}

/**
 * Displays a confirmation form to rebuild a Tugboat preview.
 *
 * @param array $form
 *   The form array.
 * @param array $form_state
 *   The form state
 * @param string $tugboat_preview_id
 *   The Tugboat preview id.
 *
 * @return array
 *   A confirmation form.
 */
function drupalorg_issue_fork_tugboat_rebuild_preview_form($form, &$form_state, $tugboat_preview_id) {
  return confirm_form(
    [
      'tugboat_preview_id' => [
        '#type' => 'value',
        '#value' => $tugboat_preview_id,
      ],
    ],
    t('Are you sure you want to rebuild the Tugboat preview?'),
    $_GET['destination'],
    t('Previews automatically rebuild when commits are pushed to their merge request. Rebuilding destroys and recreates the preview.'),
    t('Rebuild'),
    t('Cancel')
  );
}

/**
 * Form submit handler to rebuild a Tugboat preview.
 *
 * @param array $form
 *   The form array.
 * @param array $form_state
 *   The form state array.
 */
function drupalorg_issue_fork_tugboat_rebuild_preview_form_submit($form, &$form_state) {
  $preview_id = $form_state['values']['tugboat_preview_id'];
  try {
    TugboatApi::rebuildPreview($preview_id);
    drupal_set_message('The Tugboat preview has been scheduled for rebuild.');
  }
  catch (Exception $e) {
    watchdog('drupalorg_tugboat_rebuild', 'Failed to rebuild a Tugboat preview with id !preview_id. Error was !error', [
      '!preview_id' => $preview_id,
      '!error' => $e->getMessage(),
    ], WATCHDOG_ERROR);
  }
}

/**
 * Returns the build log for a given Tugboat preview.
 *
 * @param string $preview_id
 *   The Tugboat preview id.
 *
 * @return array
 *   A render array with the build log.
 */
function drupalorg_issue_fork_tugboat_preview_log($preview_id) {
  try {
    $build_log = '';
    foreach (TugboatApi::getPreviewLog($preview_id) as $build_log_line) {
      $build_log .= $build_log_line->message;
    }
    return [
      'build_log' => [
        '#type' => 'markup',
        '#markup' => '<pre class="language-bash"><code>' . check_plain($build_log) . '</code></pre>',
      ]
    ];
  }
  catch (Exception $e) {
    watchdog('drupalorg_tugboat_log', 'Failed to obtain the Tugboat build log for preview id !preview_id. Error was !error', [
      '!preview_id' => $preview_id,
      '!error' => $e->getMessage(),
    ], WATCHDOG_ERROR);
    return t('Error fetching Tugboat preview log.');
  }
}

/**
 * Menu callback, returns project members for a given gitlab project.
 */
function _drupalorg_fork_members($gitlab_project_id) {
  try {
    return drupal_json_output((bool) versioncontrol_gitlab_get_client()->api('projects')->member($gitlab_project_id, versioncontrol_gitlab_get_user_id($GLOBALS['user'])));
  }
  catch (Exception $e) {
  }

  return drupal_json_output(FALSE);
}

/**
 * Menu callback, returns additional GitLab metadata only for the URLs that are
 * migrated.
 */
function _drupalorg_gitlab_issues_metadata() {
  $client = versioncontrol_gitlab_get_client()->api('issues');
  $gitlab_base_url = variable_get('versioncontrol_gitlab_url');
  $params = drupal_get_query_parameters();
  $links = !empty($params['links']) ? $params['links'] : FALSE;
  $data = [];
  if (!empty($links) && is_array($links)) {
    $links = array_unique($links);
    foreach ($links as $link) {
      if (filter_var($link, FILTER_VALIDATE_URL)) {
        $gitlab_link = FALSE;
        $node = FALSE;
        if (strpos($link, $GLOBALS['base_url']) === 0) {
          // Internal link. Let's check if it's migrated and get the gitlab link.
          $relative_path = str_replace($GLOBALS['base_url'] . '/', '', $link);
          $system_path = drupal_lookup_path('source', $relative_path);
          if ($system_path === FALSE) {
            // Is it already a system path?
            $alias = drupal_lookup_path('alias', $relative_path);
            if ($alias) {
              $system_path = $relative_path;
            }
          }

          // Get the node and try to see if the issue is migrated.
          if ($system_path) {
            $nid = (int) str_replace('node/', '', $system_path);
            if ($nid && ($node = node_load($nid))) {
              $gitlab_link = _drupalorg_gitlab_is_issue_migrated($node, $gitlab_base_url);
            }
          }
        }
        elseif (strpos($link, $gitlab_base_url) === 0) {
          $gitlab_link = $link;
        }

        if ($gitlab_link) {
          if (preg_match('#(?<namespace>.*)/(?<project>.*)/-/issues/(?<issue>\d+)#', substr($gitlab_link, strlen($gitlab_base_url) + 1), $match)) {
            if ($gitlab_project_id = drupalorg_get_gitlab_project_id($match['namespace'], $match['project'])) {
              try {
                $gitlab_issue = $client->show($gitlab_project_id, $match['issue']);
                $data[$link] = [
                  'link' => $gitlab_link,
                  'title' => '#' . $gitlab_issue['iid'] . ' (' . $match['project'] . '): ' . $gitlab_issue['title'],
                  'status' => $gitlab_issue['state'],
                ];
              }
              catch (Exception $e) {
                $gitlab_issue = FALSE;
              }
            }
          }
        }
        elseif ($node) {
          // Issue not migrated but link might need the right text anyway.
          $data[$link] = [
            'link' => $link,
            'title' => "#$node->nid: $node->title",
            'status' => $node->field_issue_status[LANGUAGE_NONE][0]['value'],
          ];
        }
      }
    }
  }

  return drupal_json_output($data);
}

/**
 * Provides information to org owners and association staff about CPP tier
 * progress.
 *
 * @param $node
 *
 * @return $content
 */
function drupalorg_org_owner_tools(stdClass $node) {
  $case_study_count = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
    ->propertyCondition('type', 'casestudy')
    ->fieldCondition('field_case_organizations', 'target_id', $node->nid)
    ->fieldCondition('taxonomy_vocabulary_5', 'tid', [
      20236,
      194424,
      195821,
    ]) // D8, 9, 10
    ->count()->execute();

  $credits_tiers = [];
  foreach (variable_get('drupalorg_partner_level_requirements', []) as $tier => $reqs) {
    $credits_tiers[$tier] = $reqs['contribution credits'];
  }
  $highest_tier_name = array_keys($credits_tiers)[0];

  $dcp_donation_tiers = variable_get('drupalorg_certified_partner_tiers', []);
  $max_position = count($dcp_donation_tiers) - 1;
  $wrapper = entity_metadata_wrapper('node', $node);

  // Read and put dcp level into $dcp_level
  $dcp_level = NULL;
  if (!empty($wrapper->field_organization_support->value())) {
    foreach ($wrapper->field_organization_support->value() as $key => $level) {
      if (isset($dcp_donation_tiers[$level])) {
        $dcp_level = $level;
      }
    }
  }

  // DCP levels
  $position = 0;
  if (isset($dcp_level)) {
    $dcp_level_requirements = variable_get('drupalorg_partner_level_requirements');
    $dcp_dono_tiers_indexed = array_values(array_flip($dcp_donation_tiers));
    $position = array_search($dcp_level, array_keys($dcp_donation_tiers));

    if ($position < $max_position) {
      $next_position = $dcp_dono_tiers_indexed[$position + 1];
      $next_dcp_dono_tier = $dcp_donation_tiers[$next_position];
    }
  }

  // Events sponsored.
  $one_year_ago = (new DateTimeImmutable('-1 year'))->format('c');
  $query = new EntityFieldQuery();
  $events_sponsored_count = $query->fieldCondition('field_event_sponsors', 'target_id', $node->nid)
    ->fieldCondition('field_date_of_event', 'value', $one_year_ago, '>=')
    ->count()
    ->execute();

  // Supported projects.
  $view = views_get_view('drupalorg_organization_supported_projects');
  $view->preview(NULL, array($node->nid));
  $supported_project_count = count($view->result);

  // Supported contributor roles.
  $view = views_get_view('contribution_roles');
  $view_preview = $view->preview('default', [$node->nid]);
  $supported_contributor_role_count = count($view->result);

  $weights = variable_get('drupalorg_org_credit_weights');

  // Individual membership from people.
  $ind_org_memberships_count = 0;
  $result = (new EntityFieldQuery())->entityCondition('entity_type', 'field_collection_item')
    ->fieldCondition('field_organization_reference', 'target_id', $node->nid)
    ->fieldCondition('field_current', 'value', 1)->execute();
  if (!empty($result['field_collection_item'])) {
    $ind_org_memberships_count = (new EntityFieldQuery())->entityCondition('entity_type', 'user')
      ->propertyCondition('status', 1)
      ->fieldCondition('field_da_ind_membership', 'value', NULL, 'IS NOT NULL')
      ->fieldCondition('field_da_ind_membership', 'value', '', '!=')
      ->fieldCondition('field_organizations', 'value', array_keys($result['field_collection_item']))
      ->count()->execute();
  }
  $weighted_membership_rank = $weights['membership_individual'] * $ind_org_memberships_count;

  // Display content.
  drupal_set_message(t('This page is only viewable by you, the organization page’s author, and Drupal Association staff.'));
  $content['intro'] = [
    '#markup' => t('<h2>Certified partner tier information</h2><div class="partner-info"><p>Here you can see your organization’s progress on key contribution metrics, as well as an estimate of the next level you qualify. It’s important to note:</p><ul><li><strong>Each</strong> category of contribution must qualify for the next tier.</li>
    <li>Your Drupal Certified Partner level is not upgraded automatically. <a href="mailto:partnerships@association.drupal.org">Contact the partnership team</a> to upgrade.</li></ul><hr></div>'),
  ];

  // Display total credits.
  $twelve_mo_credits = empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value']) ? 0 : ceil($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value']);
  if ($twelve_mo_credits > 0) {
    $credit_tiers_rev = array_reverse($credits_tiers, TRUE);
    foreach ($credits_tiers as $key => $value) {
      if ($value < $twelve_mo_credits) {
        $qual_tier = $key;

        break;
      }
    }
    $qual_tier_credit_cutoff = $credits_tiers[$qual_tier];

    $tier_keys = array_reverse(array_keys($credits_tiers));
    $pos = array_search($qual_tier, $tier_keys);
    if ($pos !== FALSE && isset($tier_keys[$pos + 1])) {
      $next_qual_tier_cutoff = $credit_tiers_rev[$tier_keys[$pos + 1]];
    }
    else {
      $next_qual_tier_cutoff = $credit_tiers_rev[$highest_tier_name];
    }

    if (isset($dcp_level) && $wrapper->field_organization_support->value()) {
      $dcp_check_check = '✅';
    }
    else {
      $dcp_check_check = '❌';
    }

    $content['weighted_credits'] = [
      '#markup' => t('<div class="partner-tier-content"><div>' . format_plural($twelve_mo_credits, '<strong>@count</strong> Total weighted credit<br>Based on the last 12 months</div><div>The total of all weighted contribution factors. <br><br>@check Current weighted contribution credit tier would qualify you for @dcp-tier if you were a part of the program.<br>', '<strong>@count</strong> Total weighted credits<br>Based on the last 12 months</div><div>The total of all weighted contribution factors. <br><br>@check Current weighted contribution credit tier would qualify you for @dcp-tier if you were a part of the program.<br>Learn more about being a <a href="@dcp-url">Drupal Certified Partner</a>.'), [
        '@check' => $dcp_check_check,
        '@dcp-tier' => $qual_tier,
        '@year_issue_cred_weighted' => ceil($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value']),
        '@dcp-url' => url('association/become-a-drupal-certified-partner'),
      ]),
      '#suffix' => '</div></div>',
    ];

    if (($position < $max_position && $qual_tier != $highest_tier_name) && $dcp_check_check == '✅') {
      $content['weighted_credits'] = [
        '#markup' => t('<div class="partner-tier-content"><div>' . format_plural($twelve_mo_credits, '<strong>@count</strong> Total weighted credit<br>Based on the last 12 months</div><div>The total of all weighted contribution factors. <br><br>@check Current weighted contribution credit tier qualifies you for @dcp-tier.<br>', '<strong>@count</strong> Total weighted credits<br>Based on the last 12 months</div><div>The total of all weighted contribution factors.<br><br>@check Current weighted contribution credit tier is @dcp-tier.<br>Your org qualifies for @qual-tier.<br>@year_issue_cred_weighted of @next-qual-tier-credit-cutoff required to advance your qualified level.<br>'), [
          '@check' => $dcp_check_check,
          '@dcp-tier' => $dcp_donation_tiers[$dcp_level],
          '@year_issue_cred_weighted' => ceil($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value']),
          '@next_tier_credit_minimum' => $dcp_level_requirements[$next_dcp_dono_tier]['contribution credits'],
          '@next-drupal-certified-partner-tier' => $next_dcp_dono_tier,
          '@qual-tier' => $qual_tier,
          '@next-qual-tier-credit-cutoff' => $next_qual_tier_cutoff,
        ]),
        '#suffix' => '</div></div>',
      ];
    }

    elseif (($position == $max_position || $qual_tier == $highest_tier_name) && $dcp_check_check == '✅') {
      $content['weighted_credits'] = [
        '#markup' => t('<div class="partner-tier-content"><div>' . format_plural($twelve_mo_credits, '<strong>@count</strong> Total weighted credit<br>Based on the last 12 months</div><div>The total of all weighted contribution factors. <br><br><br>@check Current weighted contribution credit tier is @dcp-tier.<br>Your org qualifies for @qual-tier.<br>Thank you for being one of the top makers in Drupal.<br>', '<strong>@count</strong> Total weighted credits<br>Based on the last 12 months</div><div>The total of all weighted contribution factors.<br>@check Current weighted contribution credit tier is @dcp-tier.<br>Your org qualifies for @qual-tier.<br>@year_issue_cred_weighted of @qual-tier-credit-cutoff to maintain @qual-tier.<br>'), [
          '@check' => $dcp_check_check,
          '@dcp-tier' => $dcp_donation_tiers[$dcp_level],
          '@year_issue_cred_weighted' => ceil($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value']),
          '@qual-tier' => $qual_tier,
          '@qual-tier-credit-cutoff' => $qual_tier_credit_cutoff,
        ]),
        '#suffix' => '</div></div>',
      ];
    }
  }
  else {
    $content['weighted_credits'] = [
      '#markup' => t('<div class="partner-tier-content"><div>Your organization is not currently contributing code. <a href="@contrib-url">Learn about code contribution and get started here.</a></div><div>---</div></div>', [
        '@contrib-url' => url('community/contributor-guide/reference-information/quick-info/getting-started-contributing'),
      ]),
    ];
  }

  $content['hr'] = [
    '#markup' => '<hr>',
  ];

  // Display issues fixed.
  $credit_system = variable_get('drupalorg_credit_system', 'legacy');
  if ($credit_system === 'legacy') {
    $credit_view = views_get_view('issue_credit');
    $credit_view->set_display('organization_issue_credit_page');
    $credit_view->set_arguments([$node->nid]);
    $credit_view->pre_execute();
    $credit_view->execute();

    $year_credit_count = $credit_view->total_rows;
  }
  else {
    $year_credit_count = !empty($node->field_org_issue_credit_year) ? $node->field_org_issue_credit_year[LANGUAGE_NONE][0]['value'] : 0;
  }

  if ($year_credit_count > 0) {
    $content['issues_fixed'] = [
      '#markup' => t('<div class="partner-tier-content"><div>' . format_plural(ceil($year_credit_count), '<strong>@count</strong> issue (closed/fixed) <br>in the last 12 months<br><strong>→ @year_issue_cred_weighted weighted credit</strong></div><div><br>You have @ninety-day-weighted-credits weighted credits over the past 90 days.<p><br>Issue credits are granted by project maintainers. Contributors should follow <a href="@con-guidelines-url">org contribution guidelines</a>, <a href="@issue-ettiquette-url">issue ettiquette</a>, and the <a href="@standards-of-conduct-url">standards of conduct for the marketplace</a>. <a href="@credit-abuse-policy">Abuse of the credit system</a> can have a negative impact.</p></div></div>', '<strong>@count</strong> issues (closed/fixed) <br>in the last 12 months <br><strong>→ @year_issue_cred_weighted weighted credits</strong></div><div><br>You have @ninety-day-weighted-credits weighted credits over the past 90 days.<p><br>Issue credits are granted by project maintainers. Contributors should follow <a href="@con-guidelines-url">org contribution guidelines</a>, <a href="@issue-ettiquette-url">issue ettiquette</a>, and the <a href="@standards-of-conduct-url">standards of conduct for the marketplace</a>. <a href="@credit-abuse-policy">Abuse of the credit system</a> can have a negative impact.</p></div></div>', [
          '@year_issue_cred_weighted' => ceil(empty($node->field_weighted_org_issue_cred_yr[LANGUAGE_NONE][0]['value']) ? 0 : $node->field_weighted_org_issue_cred_yr[LANGUAGE_NONE][0]['value']),
          '@ninety-day-weighted-credits' => ceil(empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][0]['value']) ? 0 : $node->field_drupalorg_rank_components[LANGUAGE_NONE][0]['value']),
          '@con-guidelines-url' => 'https://www.youtube.com/watch?v=lu7ND0JT-8A',
          '@issue-ettiquette-url' => url('docs/develop/issues/issue-procedures-and-etiquette/issue-etiquette'),
          '@standards-of-conduct-url' => url('drupalorg/docs/marketplace/services-listing', ['fragment' => 's-standards-of-conduct-once-listed']),
          '@credit-abuse-policy' => url('drupalorg/docs/marketplace/abuse-of-the-contribution-credit-system', ['fragment' => 's-what-is-considered-abuse-of-the-contribution-credit-system']),
        ])),
    ];
  }
  else {
    $content['issues_fixed'] = [
      '#markup' => t('<div class="partner-tier-content"><div>No issues closed/fixed<br>in the past 12 months.</div><div><br>You have @ninety-day-weighted-credits weighted credits over the past 90 days.<p><br>Issue credits are granted by project maintainers. Contributors should follow <a href="@con-guidelines-url">org contribution guidelines</a>, <a href="@issue-ettiquette-url">issue ettiquette</a>, and the <a href="@standards-of-conduct-url">standards of conduct for the marketplace</a>. <a href="@credit-abuse-policy">Abuse of the credit system</a> can have a negative impact.</p></div></div>', [
        '@ninety-day-weighted-credits' => ceil(empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][0]['value']) ? 0 : $node->field_drupalorg_rank_components[LANGUAGE_NONE][0]['value']),
        '@con-guidelines-url' => 'https://www.youtube.com/watch?v=lu7ND0JT-8A',
        '@issue-ettiquette-url' => url('docs/develop/issues/issue-procedures-and-etiquette/issue-etiquette'),
        '@standards-of-conduct-url' => url('drupalorg/docs/marketplace/services-listing', ['fragment' => 's-standards-of-conduct-once-listed']),
        '@credit-abuse-policy' => url('drupalorg/docs/marketplace/abuse-of-the-contribution-credit-system', ['fragment' => 's-what-is-considered-abuse-of-the-contribution-credit-system']),
      ])
    ];
  }

  // Display credited on security advisories.
  if ($credit_system === 'legacy') {
    $credit_view = $credit_view->clone_view();
    $credit_view->set_item_option('org_sa_recent', 'filter', 'field_attribute_contribution_to_target_id', 'value', ['value' => $node->nid]);
    $credit_view->set_item_option('org_sa_recent', 'filter', 'field_for_customer_target_id', 'value', ['value' => $node->nid]);
    $credit_view->preview('org_sa_recent');
    $advisories = 0;
    foreach ($credit_view->result as $row) {
      // Display is grouped by project, ->nid is the count per-project.
      $advisories += $row->nid;
    }
  }
  else {
    // We don't keep the raw number of SA, just the calculated weighted value.
    $advisories = !empty($wrapper->field_drupalorg_rank_components[11]->value()) ? $wrapper->field_drupalorg_rank_components[11]->value() : 0;
  }

  if ($advisories > 0) {
    if ($credit_system === 'legacy') {
      $content['security_advisories'] = [
        '#markup' => t('<div class="partner-tier-content"><div>' . format_plural($advisories, '<strong>@count</strong> security advisory <br>in the last 12 months<br><strong>→ @year_sa_cred_weighted weighted credit</strong></div><div><p>Security advisory credits are granted by the <a href="/drupal-security-team">Drupal Security Team</a> for assisting with security issues.</p></div></div>', '<strong>@count</strong> security advisories <br>in the last 12 months <br><strong>→ @year_sa_cred_weighted weighted credits</strong></div><div><p>Security advisory credits are granted by the <a href="/drupal-security-team">Drupal Security Team</a> for assisting with security issues.</p></div></div>', [
          '@year_sa_cred_weighted' => ceil(empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][11]['value']) ? 0 : $node->field_drupalorg_rank_components[LANGUAGE_NONE][11]['value']),
        ])),
      ];
    }
    else {
      $content['security_advisories'] = [
        '#markup' => t('<div class="partner-tier-content"><div>Security Advisories <br>in the last 12 months<br><strong>→ @year_sa_cred_weighted weighted credit</strong></div><div><p>Security advisory credits are granted by the <a href="/drupal-security-team">Drupal Security Team</a> for assisting with security issues.</p></div></div>', [
          '@year_sa_cred_weighted' => $advisories,
        ]),
      ];
    }
  }
  else {
    $content['security_advisories'] = [
      '#markup' => t('<div class="partner-tier-content"><div>No security advisory<br>in the past 12 months.</div><div><p>Security advisory credits are granted by the <a href="/drupal-security-team">Drupal Security Team</a> for assisting with security issues.</p></div></div>'),
    ];
  }

  // Display case studies.
  if ($case_study_count) {
    $content['case_studies'] = [
      '#markup' => '<div class="partner-tier-content"><div>' . format_plural($case_study_count, '<strong>@count</strong> case study for Drupal 8 & later <br><strong>→ @case_study_rank weighted credits</strong>', '<strong>@count</strong> case studies for Drupal 8 & later <br><strong>→ @case_study_rank weighted credits</strong>', [
          '@case_study_rank' => empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][4]['value']) ? 0 : $node->field_drupalorg_rank_components[LANGUAGE_NONE][4]['value'],
        ]) . '</div><div>' . t('<p>Case studies must first be added to Drupal.org and then verified by Drupal Association staff for Certified Partner qualification.</p><a href="@case-study-submission-url" class="action-button">Add case study</a>', [
          '@case-study-submission-url' => url('node/add/casestudy'),
        ]) . '</div></div>',
    ];
  }
  else {
    $content['case_studies'] = [
      '#markup' => t('<div class="partner-tier-content"><div>No case studies for Drupal 8 or later</div><div><p>Case studies must first be added to Drupal.org and then verified by Drupal Association staff for Certified Partner qualification.</p><a href="@case-study-submission-url" class="action-button">Add case study</a>', [
          '@case-study-submission-url' => url('node/add/casestudy'),
        ]) . '</div></div>',
    ];
  }

  // Display supported projects.
  if ($supported_project_count > 0) {
    $content['drupalorg_organization_supported_projects'] = [
      '#markup' => '<div class="partner-tier-content"><div>' . format_plural($supported_project_count, '<strong>@count </strong> project supported<br><strong>→ @count weighted credits</strong>', '<strong>@count </strong> projects supported<br><strong>→ @count weighted credits</strong>') . '</div><div>' . t('Supporting a project maintainer helps keep the contributed module ecosystem healthy. Reach out to maintainers of projects you use and ask if you can provide support, they can then credit you from their project page.') . '</div></div>',
    ];
  }
  else {
    $content['drupalorg_organization_supported_projects'] = [
      '#markup' => t('<div class="partner-tier-content"><div>0 projects currently supported</div><div>Supporting a project maintainer helps keep the contributed module ecosystem healthy. Reach out to maintainers of projects you use and ask if you can provide support, they can then credit you from their project page.</div></div>'),
    ];
  }

  // Display supported contributor roles.
  if ($supported_contributor_role_count > 0) {
    $content['drupalorg_supported_contributor_roles'] = [
      '#markup' => '<div class="partner-tier-content"><div>' . format_plural($supported_contributor_role_count, '<strong>@count </strong> supported contributor role<br> <strong>→ @supported-contributor-role-credits weighted credits</strong>', '<strong>@count </strong> supported contributor roles<br> <strong>→ @supported-contributor-role-credits weighted credits</strong>', [
        '@supported-contributor-role-credits' => empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][6]['value']) ? 0 : $node->field_drupalorg_rank_components[LANGUAGE_NONE][6]['value'],
        ]) . '</div><div>' . t('Sponsoring a contributor role lets the Drupal Community thrive, whether the contributor is a mentor, documentation editor, or core maintainer, we need contributors of all kinds to support the project.') . '</div></div>',
    ];
  }
  else{
    // $supported_contributor_role_count <= 0
        $content['drupalorg_supported_contributor_roles'] = [
          '#markup' => t('<div class="partner-tier-content"><div>No credits for contributor roles</div><div>You are not currently sponsoring any contributor roles in Drupal. By sponsoring contributors you can support a healthy ecosystem.<br><a href="@role-browser-url" class="action-button">Lean more about contributor roles</a></div></div>',[
            '@role-browser-url' => url('/community/contributor-guide/find-your-role'),
          ])
        ];
  }

  // Display people on Drupal.org.
  if ($ind_org_memberships_count > 0) {
    $content['drupalorg_people_on_drupalorg'] = [
      '#markup' => t('<div class="partner-tier-content"><div><strong>@ind-mem-count</strong> individual memberships <br><strong>→ @ind-mem-rank weighted credits</strong></div><div>Making sure your team is signed up on Drupal.org shows clients the scope of your team, and gives team members an opportunity to get involved in the community. Each person on your team who is a Drupal Association member gives you a boost as well.</div></div>', [
        '@ind-mem-count' => $ind_org_memberships_count,
        '@ind-mem-rank' => $weighted_membership_rank,
      ]),
    ];
  }
  else {
    $content['drupalorg_people_on_drupalorg'] = [
      '#markup' => t('<div class="partner-tier-content"><div>No individual membership credits</div><div>Making sure your team is signed up on Drupal.org shows clients the scope of your team, and gives team members an opportunity to get involved in the community. Each person on your team who is a Drupal Association member gives you a boost as well.</div></div>'),
    ];
  }

  // Display sponsored events.
  if ($events_sponsored_count > 0) {
    $events_sponsored_credits = empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][10]['value']) ? 0 : $node->field_drupalorg_rank_components[LANGUAGE_NONE][10]['value'];
    $content['drupalorg_organization_sponsored_events'] = [
      '#markup' => '<div class="partner-tier-content"><div>' .
        format_plural($events_sponsored_count,
          '<strong>@count </strong> sponsored Drupal event in the past year',
          '<strong>@count </strong> sponsored Drupal events in the past year'
        )
        . '<br>' . t('<strong>→ @events_sponsored_credits weighted credits</strong>', [
          '@events_sponsored_credits' =>  $events_sponsored_credits,
        ])
        . '</div><div>'
        . t('Sponsoring Drupal events helps keep the Drupal ecosystem healthy.')
        . '</div></div>'
    ];
  }
  else {
    // $events_sponsored_count = 0
    $content['drupalorg_organization_sponsored_events'] = [
          '#markup' => t('<div class="partner-tier-content"><div>No credits for sponsoring Drupal events<br>in the past 12 months.</div><div>You currently have no credits for sponsoring Drupal events in the last 12 months. You can get credit for both DrupalCon sponsorship, and for sponsoring grassroots community events. They just need to list you as a sponsor on the Community Events page.<br><a href="@community-events-link" class="action-button">Community events</a></div></div>', [
            '@community-events-link' => url('/community/events'),
          ])
        ];
  }

  // Display special program.
  if (!empty($node->field_org_special_program)) {
    $content['drupalorg_org_special_program'] = [
      '#markup' => t('<div class="partner-tier-content"><div>Special programs<br><strong>→ @special-program-points weighted credits</strong></div><div>@special-program-points-description</div></div>', [
        '@special-program-points' => $node->field_org_special_program[LANGUAGE_NONE][0]['value'],
        '@special-program-points-description' => $node->field_org_special_program_desc[LANGUAGE_NONE][0]['value'],
      ]),
    ];
  }
  else {
    //$node->field_org_special_program IS empty
    $content['drupalorg_org_special_program'] =[
      '#markup' => t('<div class="partner-tier-content"><div>No credits for special programs</div><div>You currently have no credits for Drupal Association special programs. These are bespoke partnership opportunities offered by the Drupal Association to meet key goals, such as donating development support or infrastructure, or representing Drupal at external product marketing events. Contact your account manager for more information.</div></div>')
    ];
  }

  if ($credit_system === 'legacy') {
    $view = views_get_view('issue_credit');
    $content['user_credit_view'] = [
      '#prefix' => '<div class="accordion"><h3 id="org-issue-page-issue-credit-user">Organization issue credits by contributor</h3>',
      '#markup' => '<div>' . $view->preview('organization_user_issue_credit_page', [$node->nid]) . '</div>',
      '#suffix' => '</div>',
    ];
  }
  else {
    $content['user_credit_view'] = [
      '#markup' => '<div><a class="primary-button" href="/node/' . $node->nid . '/org-issue-credit-user">' . t('View this in a full page') . '</a></div>',
    ];
  }

  $content['note'] = [
    '#suffix' => t('<div class="note"><p><a href="@credit-url">How are credits weighted?</a><br>For questions or more information about your Certified Partnership level <a href="mailto:partnerships@association.drupal.org">contact the partnership team</a>.</p></div>', [
      '@credit-url' => url('contribution-credit/weights'),
    ]),
  ];

  return $content;
}

/**
 * Callback function.
 */
function _drupalorg_toggle_branch_visibility($gitlab_project_id, $branch_name, $status) {
  $branch_name = urldecode($branch_name);

  // Check token.
  if (!isset($_GET['token']) || !drupal_valid_token($_GET['token'], 'drupalorg-toggle-branch-' . $gitlab_project_id . '-' . $branch_name . '-' . $status)) {
    drupal_set_message(t('Invalid token.'), 'error');
    drupal_access_denied();
  }

  // Check that the status value is valid.
  if (!in_array($status, ['hidden', 'active'])) {
    drupal_set_message(t('Status %status is not a valid option.', ['%status' => $status]), 'error');
    drupal_goto();
  }

  // Set the visibility of the branch.
  db_merge('drupalorg_issue_fork_branches')
    ->key([
      'gitlab_project_id' => $gitlab_project_id,
      'branch' => $branch_name,
    ])
    ->fields(['status' => $status])
    ->execute();

  // Create a comment in the issue to keep track of who changed it.
  if ($fork = entity_load_single('drupalorg_issue_fork', $gitlab_project_id)) {
    drupalorg_post_issue_comment_for_credit(
      $fork->nid,
      user_load($GLOBALS['user']->uid),
      '!name changed the visibility of the branch %branch to %status.',
      [
        '%branch' => $branch_name,
        '%status' => $status,
      ]
    );
  }

  drupal_set_message(t('Branch visibility has been toggled.'));
  drupal_goto('', ['fragment' => 'main']);
}

/**
 * An interstitial page we use as a way to force validation of new accounts
 *
 * @param $node
 *
 * @return $content
 */
function drupalorg_registration_validation() {
  $account = user_load($GLOBALS['user']->uid);
  $account_wrapper = entity_metadata_wrapper('user', $account);

  // Check for FP when this page loads, just in case they successfully disabled JS blockers.
  $content['drupalorg_fpjs'] = drupal_get_form('drupalorg_form_fp_form', $account);

  // If the account fails the check, or we previously set a value indicating failure
  if (drupalorg_user_is_fingerprinted($account)) {
    drupal_goto('user');
    return;
  }

  if (!isset($_SESSION['drupalorg_registration_validation_count'])) {
    $_SESSION['drupalorg_registration_validation_count'] = 1;
  }
  else {
    $_SESSION['drupalorg_registration_validation_count'] += 1;
  }

  $content['instructions'] = [
    '#markup' => t('<p>For the health of both the community and the site we need to verify that your account is unique. Because of the configuration of your browser, we have been unable to complete the checks.</p><p>Here are some steps you can take to help with verification:</p><ul><li>Disable content blockers like uBlock Origin & Adblock Plus</li><li>Ensure JavaScript is enabled in your browser</li><li>Disable enhanced tracking protections in your browser (<a href="https://support.mozilla.org/en-US/kb/enhanced-tracking-protection-firefox-desktop#w_what-to-do-if-a-site-seems-broken" target="_blank">Firefox tracking protections</a>)</li><li><a href="@homepage">Click here to try again</a></li></ul><p>When you no longer see this page, you are welcome to re-enable content blockers and tracking protections.</p>', [
      '@homepage' => url(drupal_get_destination()['destination']),
    ]),
  ];

  if ($_SESSION['drupalorg_registration_validation_count'] > 2) {
    if ($account_wrapper->field_fingerprint->value() === 'fingerprint-failed-validation') {
      $content['support'] = ['#markup' => t('A support request has been submitted to help@drupal.org. Watch for email from that address.')];
    }
    else {
      $content['form_field'] = drupal_get_form('drupalorg_registration_validation_text_form');
    }
  }

  if (empty($account->field_fingerprint)) {
    watchdog('drupalorg_fingerprint', 'Warning, fingerprint not set.');
  }

  return $content;
}

/**
 * Form callback.
 */
function drupalorg_form_fp_form($form, $form_state) {
  $form['#attached']['js'][] = ['type' => 'setting', 'data' => [
    'drupalOrgFpjsSource' => url(drupal_get_path('module', 'drupalorg') . '/js/vendor/node_modules/@fingerprintjs/fingerprintjs-pro/dist/fp.min.js'),
    'drupalOrgFpjsParams' => variable_get('drupalorg_fpjs_params', []),
  ]];
  $form['#attached']['js'][] = drupal_get_path('module', 'drupalorg') . '/js/registration-prot.js';

  // Registration protection.
  $form['field_fingerprint'] = [
    '#title' => t('Fingerprint'),
    '#type' => 'textfield',
  ];

  $form['submit'] = [
    '#type' => 'submit',
    '#value' => t('Submit'),
  ];

  return $form;
}

/**
 * AJAX callback for drupalorg_form_fp_form().
 */
function drupalorg_form_fp_form_submit($form, &$form_state) {
  list($account) = $form_state['build_info']['args'];
  $account->field_fingerprint[LANGUAGE_NONE][0]['value'] = $form_state['values']['field_fingerprint'];
  user_save($account);

  try {
    // Get data from fingerprintjs
    $client = new GuzzleHttp\Client([
      'base_uri' => 'https://api.fpjs.io',
      'timeout' => 30.0,
    ]);

    $fpjs_response = json_decode($client->request('GET', '/visitors/' . $form_state['values']['field_fingerprint'], [
      'query' => ['limit' => '10', 'token' => variable_get('fpjs_api_prod_key')],
    ])->getbody(), TRUE);
  } catch (Exception $e) {
    // Service is down, let the user exist and check them another time later
    watchdog('drupalorg_fingerprint', 'Fingerprint.js service unavailable. visitorId: @visitorId, error: @message', [
      '@message' => $e->getMessage(),
      '@visitorId' => $account->field_fingerprint[LANGUAGE_NONE][0]['value'],
    ], WATCHDOG_ERROR);
    return;
  }

  $query = (new EntityFieldQuery())->entityCondition('entity_type', 'user')
    ->fieldCondition('field_fingerprint', 'value', $form_state['values']['field_fingerprint']);
  $count = $query->count()->execute();
  // $count > 1 means they've been here before
  // $count < 1 means tampering
  // empty($fpjs_response['visits']) checks that a valid fingerprint was sent.
  if ($count > 1 || $count < 1 || empty($fpjs_response['visits'])) {
    watchdog('drupalorg_fingerprint', 'Registration denied: reg count > 0  This fingerprint has already registered.  visitorId: @visitorId', [
      '@visitorId' => $form_state['values']['field_fingerprint'],
    ], WATCHDOG_INFO);
    $account->status = 0;

    // Send an email notifying account creator of block.
    drupal_mail('drupalorg', 'drupalorg_spam_block_notify', $account->mail, LANGUAGE_NONE, ['account' => $account]);
    drupal_set_message(t('Your Drupal.org account has been blocked by our automated spam prevention tools. If you believe this is in error, please contact us at help@drupal.org to resolve this issue. Please include your user name, browser version, operating system, and the email address you tried to to use during account creation or we may not respond.'), 'warning');

    drupalorg_append_user_note(entity_metadata_wrapper('user', $account), t('Automated block due to spam prevention duplicate account at @time', [
      '@time' => format_date(REQUEST_TIME),
    ]));

    user_save($account);
    $form_state['redirect'] = 'user/login';
  }
}

/**
 * Form callback.
 */
function drupalorg_registration_validation_text_form($form, $form_state) {

  $form['explanation'] = [
    '#prefix' => '<p>' . t('It looks like you are having trouble verifying your browser!') . '</p>',
    '#title' => t('Tell us a little bit about how you would like to use your account on Drupal.org'),
    '#type' => 'textarea',
    '#required' => TRUE,
  ];
  $form['submit'] = [
    '#type' => 'submit',
    '#value' => t('Ask for support'),
  ];

  return $form;
}

/**
 * Registration explanation form validation.
 */
function drupalorg_registration_validation_text_form_validate($form, &$form_state) {
  $explanation_length = strlen($form_state['values']['explanation']);
  $min_explanation_length = variable_get('drupalorg_minimum_fingerprint_explanation', 100);
  if($explanation_length < $min_explanation_length) {
    form_set_error('explanation', t('Please give us a longer explanation (preferably a sentence or two)'));
  }
}

/**
 * Form submit callback for drupalorg_registration_validation_text_form().
 */
function drupalorg_registration_validation_text_form_submit($form, &$form_state) {
  $account = user_load($GLOBALS['user']->uid);

  // Send mail to help@drupal.org notifying us of user intent.
  drupal_mail('drupalorg', 'drupalorg_registration_verification_failed', 'help@drupal.org', LANGUAGE_NONE, [
    'account' => $account,
    'explanation' => $form_state['values']['explanation'],
  ], $account->mail);

  $account->field_fingerprint[LANGUAGE_NONE][0]['value'] = 'fingerprint-failed-validation';
  user_save($account);

  drupal_set_message(t('Support request emailed to <a href="mailto:help@drupal.org">help@drupal.org</a>.'));
}

/**
 * Menu callback, handle Keycloak relayed event.
 *
 * @param string $event_trigger
 *   Trigger that produced the event.
 *   Either 'user' or 'admin'.
 * @param string $event_type
 *   For user events, the event type.
 *   For admin events, the resource type.
 *
 * Payload is JSON with the event data.
 *
 * E.g. a user event.
 * @code
 * {
 *   "event_trigger": "user",
 *   "event": {
 *     "id": "0c9061c8-8656-4b73-9302-9cf52e94922b",
 *     "time": "1725046809362",
 *     "type": "CODE_TO_TOKEN",
 *     "realm_id": "Main",
 *     "client_id": "account-console",
 *     "session_id": "9ecd59b9-7c17-4fc2-adc6-422f4d2bdace",
 *     "user_id": "80cefc10-1d71-4b5b-b079-39345ab5e65d",
 *     "ip_address": "192.168.1.101",
 *     "error": ""
 *   }
 * }
 * @endcode
 *
 * E.g. an admin event.
 * @code
 * {
 *   "event_trigger": "admin",
 *   "event": {
 *     "id": "4e907d65-7166-4272-aa9e-8711947514a2",
 *     "time": "1732744041989",
 *     "operation_type": "DELETE",
 *     "resource_type": "USER_SESSION",
 *     "resource_path": "sessions/8ad3019f-c52c-4a80-8769-44ade4f6ded8",
 *     "realm_id": "master",
 *     "client_id": "a35cfe7e-0d22-402c-801b-fe0d8e46e03f",
 *     "user_id": "59510008-2365-4934-878b-92ef6db21958",
 *     "ip_address": "192.168.1.10",
 *     "error": ""
 *   }
 * }
 * @endcode
 */
function drupalorg_keycloak_event($event_trigger, $event_type) {
  if (empty($event_trigger) || empty($event_type)) {
    // Missing arguments.
    watchdog('drupalorg_keycloak_event', 'Keycloak sent an event with missing arguments, event trigger was "@trigger", and event type was "@type".', [
      '@trigger' => $event_trigger,
      '@type' => $event_type,
    ], WATCHDOG_ERROR);
    drupal_add_http_header('Status', '400 Bad Request');
    print '{"error": "Empty event trigger or event type in request path"}';
    drupal_page_footer();
    exit();
  }
  $user_event_supported_types = [
    'DELETE_ACCOUNT',
    'LOGOUT',
    'LOGOUT_ERROR',
    'UPDATE_EMAIL',
    'UPDATE_PROFILE',
  ];
  $admin_event_supported_resource_types = [
    'USER',
    'USER_SESSION',
    'REALM'
  ];
  $is_supported_user_event = $event_trigger == 'user' && in_array($event_type, $user_event_supported_types);
  $is_supported_admin_event = $event_trigger == 'admin' && in_array($event_type, $admin_event_supported_resource_types);
  if (!$is_supported_user_event && !$is_supported_admin_event) {
    // Unsupported combination, stop.
    drupal_add_http_header('Status', '204 No Content');
    drupal_page_footer();
    exit();
  }
  $event = json_decode(file_get_contents('php://input'), true);
  _drupalorg_keycloak_event_handle_generic($event);
  drupal_add_http_header('Content-Type', 'text/plain');
  echo '👋';
  drupal_page_footer();
  exit();
}

/**
 * Handles KC user or admin event.
 *
 * @param array $event
 *   Parsed JSON payload on POST. See payload examples on
 *   drupalorg_keycloak_event().
 */
function _drupalorg_keycloak_event_handle_generic($event) {
  $is_logout_error = !empty($event['event']['type']) && $event['event']['type'] == 'LOGOUT_ERROR';
  if (empty($event['event']['user_id']) && !$is_logout_error) {
    // Cannot continue without a Keycloak UUID.
    watchdog('drupalorg_keycloak_event', 'Keycloak sent a user event without event.user_id set: @payload', [
      '@payload' => print_r($event, 1),
    ], WATCHDOG_ERROR);
    return;
  }
  $acting_keycloak_user_uuid = $event['event']['user_id'];

  switch ($event['event_trigger']) {
    case 'user':
      $keycloak_uuid = $acting_keycloak_user_uuid;
      $event_type = $event['event']['type'];
      if (in_array($event_type, ['LOGOUT', 'LOGOUT_ERROR'])) {
        // Only log and return for logout related events.
        if ($event_type == 'LOGOUT') {
          watchdog('drupalorg_keycloak_event', 'Keycloak user @uuid logged out.', [
            '@uuid' => $keycloak_uuid,
          ], WATCHDOG_INFO);
          return;
        }
        watchdog('drupalorg_keycloak_event', 'Keycloak user @uuid logout failed: @payload', [
          '@uuid' => $keycloak_uuid,
          '@payload' => print_r($event, 1),
        ], WATCHDOG_INFO);
        return;
      }
      break;

    case 'admin':
      if ($acting_keycloak_user_uuid === '710734d3-5383-47b0-bcf6-95d989096a4b') {
        // Do not process service-account-initiated updates, because that would
        // create an infinite loop.
        return;
      }
      if ($event['event']['resource_type'] == 'USER') {
        $keycloak_uuid = preg_replace('#^users/#', '', $event['event']['resource_path']);
        $event_type = $event['event']['operation_type'] . '-' . $event['event']['resource_type'];
      }
      else {
        // Only log and return other resource types.
        // TODO Maybe translate event structure into a more detailed message.
        watchdog('drupalorg_keycloak_event', 'Keycloak admin event by @uuid: @payload', [
          '@uuid' => $acting_keycloak_user_uuid,
          '@payload' => print_r($event, 1),
        ], WATCHDOG_INFO);
        return;
      }
      break;

    default:
      // Unknown case, log and exit.
      watchdog('drupalorg_keycloak_event', 'Keycloak sent an invalid event trigger @trigger: @payload', [
        '@trigger' => $trigger,
        '@payload' => print_r($event, 1),
      ], WATCHDOG_ERROR);
      return;
  }

  // Either a user event or an admin event of resource type user, where
  // $keycloak_uuid is the user that the action is being taken on.
  $uid = db_query("SELECT uid FROM {authmap} WHERE module = 'openid_connect_keycloak' AND authname = :authname", [
    ':authname' => $keycloak_uuid,
  ])->fetchField();
  if (empty($uid)) {
    // Cannot find a related user id.
    watchdog('drupalorg_keycloak_event', 'Provided Keycloak UUID "@uuid" does not have a related entry on authmap: @payload', [
      '@uuid' => print_r($keycloak_uuid, TRUE),
      '@payload' => print_r($event, TRUE),
    ], WATCHDOG_ERROR);
    return;
  }

  drupal_register_shutdown_function([DrupalQueue::get('drupalorg_keycloak_event'), 'createItem'], [
    'uid' => $uid,
    'keycloak_uuid' => $keycloak_uuid,
    'keycloak_event_type' => $event_type,
  ]);
}

/**
 * Menu callback.
 */
function _drupalorg_user_contribution_records_by_project($username, $project_id) {
  $project = node_load($project_id);
  if (!$project || !project_node_is_project($project)) {
    drupal_set_message(t('Project not found.'));
    return drupal_not_found();
  }

  $user = user_load_by_name($username);
  if (!$user) {
    drupal_set_message(t('User not found.'));
    return drupal_not_found();
  }

  $data = [];
  $cache_id = 'drupalorg_contribution_record_credits_by_user_organized_by_project:' . md5($user->name) . ':' . $project_id;
  if (($cache = cache_get($cache_id))) {
    $data = $cache->data;
  }
  else {
    // Machine name in the endpoint is the one where the code is.
    $wrapper = entity_metadata_wrapper('node', $project);
    $repo = versioncontrol_project_repository_load($project->nid);
    $machine_name = !empty($repo->name) ? $repo->name : $wrapper->field_project_machine_name->value();

    $username = urlencode($username);
    $url = '/contribution-records-by-user?username=' . $username . '&machine_name=' . $machine_name;
    $result = drupalorg_make_contribution_records_jsonapi_request($url, TRUE);
    if (empty($result['message']) && $result['data']) {
      $data = _drupalorg_contribution_records_organized_by_project($result['data']);
    }

    $cache_minutes = variable_get('drupalorg_contribution_record_credits_cache_minutes', 60);
    cache_set($cache_id, $data, 'cache', REQUEST_TIME + $cache_minutes * 60);
  }

  $issue_list = !empty($data[$project_id]) ? $data[$project_id]['issues'] : [];
  $issue_list_links = [];
  foreach ($issue_list as $issue_info) {
    $issue_list_links[] =
      l($issue_info['title'], $issue_info['url']) . ' ' .
      t('updated @date', ['@date' => format_date(strtotime($issue_info['datetime']), 'medium')]);
  }

  // Build the page.
  drupal_set_title(t('@project issues credited to @user', [
    '@project' => $project->title,
    '@user' => $username,
  ]));
  $back_link = '<p>◀︎ ' . l(t('Back to @user\'s profile', ['@user' => $username]), 'u/' . $username) . '</p>';
  // theme('item_list'...) adds classes that break the styling, so build the list manually.
  $issues_markup = !empty($issue_list_links) ? '<ul><li>' . implode('</li><li>', $issue_list_links) . '</li></ul>' : t('No contribution records were found.');

  return [
    'back_link' => ['#markup' => $back_link],
    'content' => ['#markup' => $issues_markup],
  ];
}

/**
 * Menu callback.
 */
function _drupalorg_user_contribution_records($username, $type = 'issues') {
  $user = user_load_by_name($username);
  if (!$user) {
    drupal_set_message(t('User not found.'));
    return drupal_not_found();
  }

  if (!in_array($type, ['issues', 'security_advisories'])) {
    drupal_set_message(t('Type not valid.'));
    return drupal_not_found();
  }

  $params = drupal_get_query_parameters();
  $months = !empty($params['months']) ? (int) $params['months'] : 0;
  $page = !empty($params['page']) ? (int) $params['page'] : 0;
  $data = [];

  $cache_id = 'drupalorg_contribution_record_credits_by_user_last_months:' . md5($user->name) . ':' . $months . ':' . $page;
  if (($cache = cache_get($cache_id))) {
    $data = $cache->data;
  }
  else {
    $username = urlencode($username);
    $url = '/contribution-records-by-user?username=' . $username . '&months=' . $months . '&page=' . $page;
    $result = drupalorg_make_contribution_records_jsonapi_request($url);
    if (empty($result['message']) && $result['data']) {
      $data = [
        'data' => _drupalorg_contribution_records_organized_by_project($result['data']),
        'total' => $result['total']
      ];
    }

    $cache_minutes = variable_get('drupalorg_contribution_record_credits_cache_minutes', 60);
    cache_set($cache_id, $data, 'cache', REQUEST_TIME + $cache_minutes * 60);
  }
  if (empty($data['data'])) {
    $data['data'] = [];
  }

  // Build the page.
  drupal_set_title(t('@type credited to @user', [
    '@user' => $username,
    '@type' => ($type == 'issues') ? 'Issues' : 'Security Advisories'
  ]));
  $back_link = '<p>◀︎ ' . l(t('Back to @user\'s profile', ['@user' => $username]), 'u/' . $username) . '</p>';

  $issue_counter = 0;
  $issues_markup = [];
  foreach ($data['data'] as $project_id => $project_contribution_records) {
    $issue_list_links = [];
    foreach ($project_contribution_records[$type] as $issue_info) {
      $issue_counter++;
      $issue_list_links[] =
        l($issue_info['title'], $issue_info['url']) . ' ' .
        t('updated @date', ['@date' => format_date(strtotime($issue_info['datetime']), 'medium')]);
    }
    $issues_markup[$project_contribution_records['title']] = '<h3>' . l($project_contribution_records['title'], $project_contribution_records['url']) . '</h3>';
    // theme('item_list'...) adds classes that break the styling, so build the list manually.
    $issues_markup[$project_contribution_records['title']] .= !empty($issue_list_links) ? '<ul><li>' . implode('</li><li>', $issue_list_links) . '</li></ul>' : '';
  }

  if (empty($issues_markup)) {
    $issues_markup = t('No credited @type found', [
      '@type' => ($type == 'issues') ? 'issues' : 'Security Advisories',
    ]);
  }
  else {
    ksort($issues_markup);
    $issues_markup = implode(PHP_EOL, $issues_markup);
  }

  $range_form = drupal_get_form('_drupalorg_date_range_options_form');
  // As we are using method='get' we don't want/need the following params to show up.
  unset($range_form['form_id']);
  unset($range_form['form_token']);
  unset($range_form['form_build_id']);

  $page_size = variable_get('drupalorg_contribution_record_credits_page_size', 50);
  $from = $page * $page_size + 1;
  $to = $from + ($issue_counter - 1);
  $total = !empty($data['total']) ? $data['total'] : 0;
  $summary = $total ? t('<p>Displaying @from - @to of @total - sorted by last updated</p>', [
    '@total' => $total,
    '@from' => $from,
    '@to' => $to,
  ]) : '';
  pager_default_initialize($total, $page_size);

  return [
    'back_link' => ['#markup' => $back_link],
    'summary' => ['#markup' => $summary],
    'range_form' => $range_form,
    'content' => ['#markup' => $issues_markup],
    'pager' => ['#markup' => theme('pager')]
  ];
}

/**
 * Date range form.
 */
function _drupalorg_date_range_options_form($form, &$form_state, $hide_submit = TRUE, $inline_elements = FALSE, $months = TRUE, $username = FALSE, $machine_name = FALSE) {
  $form['#method'] = 'get';

  $params = drupal_get_query_parameters();
  if ($months) {
    $form['months'] = [
      '#title' => t('Date range'),
      '#type' => 'select',
      '#default_value' => !empty($params['months']) ? (int) $params['months'] : 0,
      '#options' => [
        0 => t('- Any -'),
        12 => t('Past year'),
        3 => t('Past 90 days'),
        2 => t('Past 60 days'),
        1 => t('Past 30 days'),
      ],
    ];
    if ($hide_submit) {
      $form['months']['#attributes'] = [
        'onChange' => 'this.form.submit();',
      ];
    }
  }
  if ($username) {
    $form['username'] = [
      '#title' => t('Contributor'),
      '#type' => 'textfield',
      '#description' => t('Use the username'),
      '#attributes' => [
        'size' => 20,
      ],
      '#default_value' => !empty($params['username']) ? $params['username'] : '',
    ];
    if ($hide_submit) {
      $form['username']['#attributes']['onChange'] = 'this.form.submit();';
    }
  }
  if ($machine_name) {
    $form['machine_name'] = [
      '#title' => t('Repository name'),
      '#type' => 'textfield',
      '#description' => t('eg: admin_toolbar'),
      '#attributes' => [
        'size' => 20,
      ],
      '#default_value' => !empty($params['machine_name']) ? $params['machine_name'] : '',
    ];
    if ($hide_submit) {
      $form['machine_name']['#attributes']['onChange'] = 'this.form.submit();';
    }
  }

  $form['submit'] = [
    '#type' => 'submit',
    '#value' => t('Apply'),
    '#attributes' => [
      'class' => [($hide_submit ? 'js-hide': '')],
    ],
  ];

  if ($inline_elements) {
    $form['#attributes'] = [
      'class' => ['views-exposed-form', 'clearfix'],
    ];
    if ($months) {
      $form['months']['#prefix'] = '<div class="views-exposed-widget">';
      $form['months']['#suffix'] = '</div>';
    }
    if ($username) {
      $form['username']['#prefix'] = '<div class="views-exposed-widget">';
      $form['username']['#suffix'] = '</div>';
    }
    if ($machine_name) {
      $form['machine_name']['#prefix'] = '<div class="views-exposed-widget">';
      $form['machine_name']['#suffix'] = '</div>';
    }
    $form['submit']['#prefix'] = '<div class="views-exposed-widget">';
    $form['submit']['#suffix'] = '</div>';
  }

  return $form;
}

/**
 * Menu callback.
 */
function _drupalorg_organization_contribution_records_by_project($organization, $project_id) {
  $project = node_load($project_id);
  if (!$project || !project_node_is_project($project)) {
    drupal_set_message(t('Project not found.'));
    return drupal_not_found();
  }

  $data = [];
  $cache_id = 'drupalorg_contribution_record_credits_by_organization_organized_by_project:' . $organization->nid . ':' . $project_id;
  if (($cache = cache_get($cache_id))) {
    $data = $cache->data;
  }
  else {
    // Machine name in the endpoint is the one where the code is.
    $wrapper = entity_metadata_wrapper('node', $project);
    $repo = versioncontrol_project_repository_load($project->nid);
    $machine_name = !empty($repo->name) ? $repo->name : $wrapper->field_project_machine_name->value();

    $url = '/contribution-records-by-organization?organization=' . urlencode($organization->title) . '&machine_name=' . $machine_name;
    $result = drupalorg_make_contribution_records_jsonapi_request($url, TRUE);
    if (empty($result['message']) && $result['data']) {
      $data = _drupalorg_contribution_records_organized_by_project($result['data']);
    }

    $cache_minutes = variable_get('drupalorg_contribution_record_credits_cache_minutes', 60);
    cache_set($cache_id, $data, 'cache', REQUEST_TIME + $cache_minutes * 60);
  }

  $issue_list = !empty($data[$project_id]) ? $data[$project_id]['issues'] : [];
  $issue_list_links = [];
  foreach ($issue_list as $issue_info) {
    $issue_list_links[] =
      l($issue_info['title'], $issue_info['url']) . ' ' .
      t('updated @date', ['@date' => format_date(strtotime($issue_info['datetime']), 'medium')]);
  }

  // Build the page.
  drupal_set_title(t('@project issues credited to @organization', [
    '@project' => $project->title,
    '@organization' => $organization->title,
  ]));
  $back_link = '<p>◀︎ ' . l(t('Back to @organization\'s profile', ['@organization' => $organization->title]), 'node/' . $organization->nid) . '</p>';
  // theme('item_list'...) adds classes that break the styling, so build the list manually.
  $issues_markup = !empty($issue_list_links) ? '<ul><li>' . implode('</li><li>', $issue_list_links) . '</li></ul>' : t('No contribution records were found.');

  return [
    'back_link' => ['#markup' => $back_link],
    'content' => ['#markup' => $issues_markup],
  ];
}

/**
 * Menu callback.
 */
function _drupalorg_organization_contribution_records_by_user_export($organization) {
  return _drupalorg_organization_contribution_records_by_user($organization, 'csv');
}

/**
 * Menu callback.
 */
function _drupalorg_organization_contribution_records_by_user($organization, $output_as = 'page') {
  if (
    empty($organization) ||
    !in_array($output_as, ['csv', 'page']) ||
    $organization->type !== 'organization'
  ) {
    return drupal_not_found();
  }

  $parameters = drupal_get_query_parameters();
  $parameters_md5 = !empty($parameters) ? md5(json_encode($parameters)) : 'empty';
  $query = [
    'organization' => urlencode($organization->title),
    'username' => !empty($parameters['username']) ? $parameters['username'] : NULL,
    'machine_name' => !empty($parameters['machine_name']) ? $parameters['machine_name'] : NULL,
    'months' => !empty($parameters['months']) ? (int) $parameters['months'] : NULL,
    'page' => !empty($parameters['page']) ? (int) $parameters['page'] : 0,
  ];

  $data = [];
  $cache_id = 'drupalorg_contribution_record_credits_by_organization_organized_by_users:' . $organization->nid . ':' . $parameters_md5 . ':' . $output_as;
  if (($cache = cache_get($cache_id))) {
    $data = $cache->data;
  }
  else {
    $url = '/contribution-records-by-organization-by-user?' . http_build_query($query);
    $result = drupalorg_make_contribution_records_jsonapi_request($url);
    if (empty($result['message']) && $result['data']) {
      $data = _drupalorg_contribution_records_organized_by_user($result);
    }

    $cache_minutes = variable_get('drupalorg_contribution_record_credits_cache_minutes', 60);
    cache_set($cache_id, $data, 'cache', REQUEST_TIME + $cache_minutes * 60);
  }

  if ($output_as === 'csv') {
    drupal_add_http_header('Content-Type', 'text/csv; utf-8');
    drupal_add_http_header('Content-Disposition', 'attachment;filename=org-credits-export.csv');

    $output = '';
    $keys = [
      // t('Username'),
      t('User link'),
      t('Issue title'),
      t('Issue link'),
      t('Date'),
      t('Project link'),
    ];
    $output .= implode("\t", $keys) . "\n";
    if (!empty($data)) {
      foreach ($data as $user_id => $issue_list) {
        // $user = user_load($user_id);
        $user_link = url('user/' . $user_id, ['absolute' => TRUE]);
        foreach ($issue_list as $issue_info) {
          $row = [
            // $user->name,
            $user_link,
            $issue_info['title'],
            $issue_info['url'],
            format_date(strtotime($issue_info['datetime']), 'medium'),
            !empty($issue_info['project_link']) ? $issue_info['project_link'] : '',
          ];
          $output .= implode("\t", $row) . "\n";
        }
      }
    }
    ob_clean();
    print $output;
    exit;
  }

  $issue_counter = 0;
  $issue_list_links = [];
  if (!empty($data)) {
    foreach ($data as $user_id => $issue_list) {
      foreach ($issue_list as $issue_info) {
        $issue_counter++;
        $title = $issue_info['title'];
        $issue_link = l($issue_info['title'], $issue_info['url']) . ' ' . t('updated @date', ['@date' => format_date(strtotime($issue_info['datetime']), 'medium')]);
        if (!empty($issue_info['project_name'])) {
          $project_link = !empty($issue_info['project_link']) ? l($issue_info['project_name'], $issue_info['project_link']) : $issue_info['project_name'];
          $issue_link = $project_link . ': ' . $issue_link;
        }

        $issue_list_links[$user_id][] = $issue_link;
      }
    }
  }

  // Filter form.
  $filter_form = drupal_get_form('_drupalorg_date_range_options_form', FALSE, TRUE, TRUE, TRUE, TRUE);
  // As we are using method='get' we don't want/need the following params to show up.
  unset($filter_form['form_id']);
  unset($filter_form['form_token']);
  unset($filter_form['form_build_id']);

  // Pagination.
  $page_size = variable_get('drupalorg_contribution_record_credits_page_size', 50);
  $from = $query['page'] * $page_size + 1;
  $to = $from + ($issue_counter - 1);
  $total = !empty($data['total']) ? $data['total'] : 0;
  $summary = $total ? t('<p>Displaying @from - @to of @total - sorted by contributor and then last updated</p>', [
    '@total' => $total,
    '@from' => $from,
    '@to' => $to,
  ]) : '';
  pager_default_initialize($total, $page_size);

  $notice = '<div>' . t('<b>Please note:</b> this view organizes the projects and issues which people have attributed to your organization for credit. You may see duplicate issues in this list when multiple people at your organization worked on the issue. Each issue is only counted once for organization credit, regardless of how many people contributed to it.') . '</div><br>';

  $issues_markup = '';
  if (empty($issue_list_links)) {
    $issues_markup = '<div>' . t('No contribution records were found.') . '</div>';
  }
  else {
    foreach ($issue_list_links as $user_id => $links) {
      $user = user_load($user_id);
      $user_link = l($user->name, 'user/' . $user->uid);
      $issues_markup .= '<div class="user-credits"><h3>' . $user_link . '</h3>';
      $issues_markup .= '<ul><li>' . implode('</li><li>', $links) . '</li></ul>';
      $issues_markup .= '</div>';
    }
  }

  $query = http_build_query($parameters);
  $csv_button = '<div class="feed-icon">
    <a href="/node/' . $organization->nid . '/org-issue-credit-user-export?' . $query . '">
      <img src="/sites/all/modules/views_data_export/images/csv.png" alt="CSV Export" title="CSV Export">
    </a>
  </div>';
  $issues_markup .= $csv_button;

  // Build the page.
  drupal_set_title(t('Issues credited to @organization', [
    '@organization' => $organization->title,
  ]));
  if (drupalorg_organization_manage_access($organization)) {
    $back_link = '<p>◀︎ ' . l(t('Back to @organization\'s owner tools', ['@organization' => $organization->title]), 'node/' . $organization->nid . '/tools') . '</p>';
  }
  else {
    $back_link = '<p>◀︎ ' . l(t('Back to @organization\'s profile', ['@organization' => $organization->title]), 'node/' . $organization->nid) . '</p>';
  }
  return [
    'back_link' => ['#markup' => $back_link],
    'summary' => ['#markup' => $summary],
    'notice' => ['#markup' => $notice],
    'filter_form' => $filter_form,
    'content' => ['#markup' => $issues_markup],
    'pager' => ['#markup' => theme('pager')]
  ];
}

/**
 * Menu callback.
 */
function _drupalorg_organization_contribution_records($organization, $type = 'issues') {
  if ($organization->type != 'organization') {
    drupal_set_message(t('Not an organization.'));
    return drupal_not_found();
  }

  if (!in_array($type, ['issues', 'security_advisories'])) {
    drupal_set_message(t('Type not valid.'));
    return drupal_not_found();
  }

  $params = drupal_get_query_parameters();
  $months = !empty($params['months']) ? (int) $params['months'] : 0;
  $page = !empty($params['page']) ? (int) $params['page'] : 0;
  $data = [];

  $cache_id = 'drupalorg_contribution_record_credits_by_organization_last_months:' . md5($organization->title) . ':' . $months . ':' . $page;
  if (($cache = cache_get($cache_id))) {
    $data = $cache->data;
  }
  else {
    $url = '/contribution-records-by-organization?organization=' . urlencode($organization->title) . '&months=' . $months . '&page=' . $page;
    $result = drupalorg_make_contribution_records_jsonapi_request($url);
    if (empty($result['message']) && $result['data']) {
      $data = [
        'data' => _drupalorg_contribution_records_organized_by_project($result['data']),
        'total' => $result['total']
      ];
    }

    $cache_minutes = variable_get('drupalorg_contribution_record_credits_cache_minutes', 60);
    cache_set($cache_id, $data, 'cache', REQUEST_TIME + $cache_minutes * 60);
  }
  if (empty($data['data'])) {
    $data['data'] = [];
  }

  // Build the page.
  drupal_set_title(t('@type credited to @organization', [
    '@organization' => $organization->title,
    '@type' => ($type == 'issues') ? 'Issues' : 'Security Advisories',
  ]));
  $back_link = '<p>◀︎ ' . l(t('Back to @organization\'s profile', ['@organization' => $organization->title]), 'node/' . $organization->nid) . '</p>';

  $issue_counter = 0;
  $issues_markup = [];
  foreach ($data['data'] as $project_id => $project_contribution_records) {
    $issue_list_links = [];
    foreach ($project_contribution_records[$type] as $issue_info) {
      $issue_counter++;
      $issue_list_links[] =
        l($issue_info['title'], $issue_info['url']) . ' ' .
        t('updated @date', ['@date' => format_date(strtotime($issue_info['datetime']), 'medium')]);
    }
    $issues_markup[$project_contribution_records['title']] = '<h3>' . l($project_contribution_records['title'], $project_contribution_records['url']) . '</h3>';
    // theme('item_list'...) adds classes that break the styling, so build the list manually.
    $issues_markup[$project_contribution_records['title']] .= !empty($issue_list_links) ? '<ul><li>' . implode('</li><li>', $issue_list_links) . '</li></ul>' : '';
  }

  if (empty($issues_markup)) {
    $issues_markup = t('No credited @type found', [
      '@type' => ($type == 'issues') ? 'issues' : 'Security Advisories'
    ]);
  }
  else {
    ksort($issues_markup);
    $issues_markup = implode(PHP_EOL, $issues_markup);
  }

  $range_form = drupal_get_form('_drupalorg_date_range_options_form');
  // As we are using method='get' we don't want/need the following params to show up.
  unset($range_form['form_id']);
  unset($range_form['form_token']);
  unset($range_form['form_build_id']);

  $page_size = variable_get('drupalorg_contribution_record_credits_page_size', 50);
  $from = $page * $page_size + 1;
  $to = $from + ($issue_counter - 1);
  $total = !empty($data['total']) ? $data['total'] : 0;
  $summary = $total ? t('<p>Displaying @from - @to of @total - sorted by last updated</p>', [
    '@total' => $total,
    '@from' => $from,
    '@to' => $to,
  ]) : '';
  pager_default_initialize($total, $page_size);

  return [
    'back_link' => ['#markup' => $back_link],
    'summary' => ['#markup' => $summary],
    'range_form' => $range_form,
    'content' => ['#markup' => $issues_markup],
    'pager' => ['#markup' => theme('pager')]
  ];
}
